# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import difflib
from dataclasses import asdict
from pathlib import Path
from textwrap import dedent, indent
from typing import Union

import yaml

from fixit.common.base import LintConfig, LintRuleT
from fixit.common.config import get_rules_from_config
from fixit.common.utils import (
    DEFAULT_CONFIG,
    DEFAULT_FILENAME,
    InvalidTestCase,
    ValidTestCase,
)


def _add_code_indent(code: str) -> str:
    return indent(code, "    ") + "\n"


def _add_reference_label(label: str) -> str:
    return f".. _{label}:\n"


def _add_title_style(title: str, symbol: str) -> str:
    title_length = len(title)
    return dedent(
        f"""
            {symbol * title_length}
            {title}
            {symbol * title_length}
            """
    )


def _add_config(config: LintConfig) -> str:
    s = ""
    if config != DEFAULT_CONFIG:
        conf = asdict(config)
        default_conf = asdict(DEFAULT_CONFIG)
        custom_conf = {
            key: val for key, val in conf.items() if val != default_conf[key]
        }
        s = dedent(
            """
            config:

            .. code-block:: yaml

            """
        )
        s += _add_code_indent(
            yaml.dump(custom_conf, indent=2, default_flow_style=False)
        )
    return s


def _get_example(example: Union[ValidTestCase, InvalidTestCase], index: int) -> str:
    s = ""
    source = dedent(example.code)
    filename = (
        f"path: :file:`{example.filename}`"
        if example.filename != DEFAULT_FILENAME
        else ""
    )
    case = dedent(
        f"""
        # {index + 1}:

        {_add_code_indent(_add_code_indent(_add_config(example.config)))}

        {filename}

        .. code-block:: python

        {_add_code_indent(_add_code_indent(_add_code_indent(source)))}
        """
    )
    s += case
    if isinstance(example, InvalidTestCase):
        replacement = example.expected_replacement
        if replacement is not None:
            autofix = dedent(replacement)
            diff = "\n".join(
                difflib.unified_diff(
                    source.splitlines(), autofix.splitlines(), lineterm=""
                )
            )
            s += dedent(
                """
                Autofix:

                .. code-block:: python

                """
            )
            s += _add_code_indent(diff)
    return s


def gen_example_cases(rule: LintRuleT, key: str, to_fold_examples: bool = True) -> str:
    s = ""
    examples_before_folding = 3
    if hasattr(rule, key):
        s += _add_title_style(f"{key} Code Examples", "-")
        for idx, example in enumerate(getattr(rule, key)):
            if to_fold_examples and idx >= examples_before_folding:
                if idx == examples_before_folding:
                    s += "\n.. container:: toggle\n\n"
                s += _add_code_indent(_get_example(example, idx))
            else:
                s += _get_example(example, idx)

    return s


def _has_autofix(rule: LintRuleT) -> bool:
    invalids = getattr(rule, "INVALID", [])
    return any(i.expected_replacement for i in invalids)


def create_rule_doc(directory: Path, to_fold_examples: bool = False) -> None:
    directory.mkdir(exist_ok=True)
    for rule in get_rules_from_config():
        rule_name = rule.__name__
        with (directory / f"{rule_name}.rst").open("w") as fp:
            fp.write(_add_reference_label(rule_name))
            fp.write(_add_title_style(rule_name, "="))
            doc = rule.__doc__
            if doc is not None:
                fp.write(dedent(doc))
            message = getattr(rule, "MESSAGE", None)
            if message is not None:
                fp.write(_add_title_style("Message", "-"))
                fp.write(message + "\n")
            fp.write(
                _add_title_style(
                    f"Has Autofix: {'Yes' if _has_autofix(rule) else 'No'}", "-"
                )
            )
            fp.write(gen_example_cases(rule, "VALID", to_fold_examples))
            fp.write(gen_example_cases(rule, "INVALID", to_fold_examples))



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import contextlib
import io
import json
import os
import tempfile
from typing import Any, Dict

from libcst.testing.utils import UnitTest

from fixit.cli import run_ipc
from fixit.cli.args import LintWorkers
from fixit.cli.tests.test_lint_opts import generate_mock_lint_opt


EXPECTED_SUCCESS_REPORT: Dict[str, Any] = json.loads(
    """{"path": "fill-this-out", "status": "success", "reports": ["fake picklable report"]}"""
)
EXPECTED_FAILURE_REPORT: Dict[str, Any] = json.loads(
    """{"path": "fill-this-out", "status": "failure", "reports": ["fake picklable report"]}"""
)


class IpcTest(UnitTest):
    def setUp(self) -> None:
        self.opts = generate_mock_lint_opt()

    def test_single_path_ipc(self) -> None:
        with io.StringIO() as buffer, tempfile.TemporaryDirectory() as prefix, contextlib.redirect_stdout(
            buffer
        ):
            # create a valid file for the test to run against
            path = "path.py"

            with open(os.path.join(prefix, path), "w") as fd:
                fd.write("""test_str = 'hello world'""")

            run_ipc(
                opts=self.opts,
                paths=[path],
                prefix=prefix,
                workers=LintWorkers.USE_CURRENT_THREAD,
            )

            # get values from the buffer before we close it
            buffer.flush()
            output = buffer.getvalue()

        report = json.loads(output)

        target_report = EXPECTED_SUCCESS_REPORT.copy()
        target_report["path"] = os.path.join(prefix, path)

        self.assertDictEqual(report, target_report)

    def test_multi_path_ipc(self) -> None:
        with io.StringIO() as buffer, tempfile.TemporaryDirectory() as prefix, contextlib.redirect_stdout(
            buffer
        ):
            path_a = "path_a.py"
            path_b = "path_b.py"
            # this path doesn't exist at all, but the runner should still handle it gracefully
            path_c = "does_not_exist.tmp"

            # create a valid file for the test to run against
            with open(os.path.join(prefix, path_a), "w") as fd_a:
                fd_a.write("""test_str = 'hello world'""")

            # now create an invalid one
            # mismatched tab-indent will do the trick
            with open(os.path.join(prefix, path_b), "w") as fd_b:
                fd_b.write("""\ta = 1\nb = 2""")

            run_ipc(
                opts=self.opts,
                paths=[path_a, path_b, path_c],
                prefix=prefix,
                workers=LintWorkers.USE_CURRENT_THREAD,
            )

            # get values from the buffer before we close it
            buffer.flush()
            output = buffer.getvalue()

        # each report is separated by a new-line
        reports = output.strip().split("\n")
        self.assertEqual(len(reports), 3)
        report_a, report_b, report_c = [json.loads(report) for report in reports]

        target_report_a = EXPECTED_SUCCESS_REPORT.copy()
        target_report_a["path"] = os.path.join(prefix, path_a)

        target_report_b = EXPECTED_FAILURE_REPORT.copy()
        target_report_b["path"] = os.path.join(prefix, path_b)

        target_report_c = EXPECTED_FAILURE_REPORT.copy()
        target_report_c["path"] = os.path.join(prefix, path_c)

        self.assertDictEqual(report_a, target_report_a)
        self.assertDictEqual(report_b, target_report_b)
        self.assertDictEqual(report_c, target_report_c)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import libcst as cst
import libcst.matchers as m
from libcst._nodes.whitespace import SimpleWhitespace
from libcst.metadata import QualifiedName, QualifiedNameProvider, QualifiedNameSource

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


class ExplicitFrozenDataclassRule(CstLintRule):
    """
    Encourages the use of frozen dataclass objects by telling users to specify the
    kwarg.

    Without this lint rule, most users of dataclass won't know to use the kwarg, and
    may unintentionally end up with mutable objects.
    """

    MESSAGE: str = (
        "When using dataclasses, explicitly specify a frozen keyword argument. "
        + "Example: `@dataclass(frozen=True)` or `@dataclass(frozen=False)`. "
        + "Docs: https://docs.python.org/3/library/dataclasses.html"
    )
    METADATA_DEPENDENCIES = (QualifiedNameProvider,)
    VALID = [
        Valid(
            """
            @some_other_decorator
            class Cls: pass
            """
        ),
        Valid(
            """
            from dataclasses import dataclass
            @dataclass(frozen=False)
            class Cls: pass
            """
        ),
        Valid(
            """
            import dataclasses
            @dataclasses.dataclass(frozen=False)
            class Cls: pass
            """
        ),
        Valid(
            """
            import dataclasses as dc
            @dc.dataclass(frozen=False)
            class Cls: pass
            """
        ),
        Valid(
            """
            from dataclasses import dataclass as dc
            @dc(frozen=False)
            class Cls: pass
            """
        ),
    ]
    INVALID = [
        Invalid(
            """
            from dataclasses import dataclass
            @some_unrelated_decorator
            @dataclass  # not called as a function
            @another_unrelated_decorator
            class Cls: pass
            """,
            line=3,
            expected_replacement="""
            from dataclasses import dataclass
            @some_unrelated_decorator
            @dataclass(frozen=True)  # not called as a function
            @another_unrelated_decorator
            class Cls: pass
            """,
        ),
        Invalid(
            """
            from dataclasses import dataclass
            @dataclass()  # called as a function, no kwargs
            class Cls: pass
            """,
            line=2,
            expected_replacement="""
            from dataclasses import dataclass
            @dataclass(frozen=True)  # called as a function, no kwargs
            class Cls: pass
            """,
        ),
        Invalid(
            """
            from dataclasses import dataclass
            @dataclass(other_kwarg=False)
            class Cls: pass
            """,
            line=2,
            expected_replacement="""
            from dataclasses import dataclass
            @dataclass(other_kwarg=False, frozen=True)
            class Cls: pass
            """,
        ),
        Invalid(
            """
            import dataclasses
            @dataclasses.dataclass
            class Cls: pass
            """,
            line=2,
            expected_replacement="""
            import dataclasses
            @dataclasses.dataclass(frozen=True)
            class Cls: pass
            """,
        ),
        Invalid(
            """
            import dataclasses
            @dataclasses.dataclass()
            class Cls: pass
            """,
            line=2,
            expected_replacement="""
            import dataclasses
            @dataclasses.dataclass(frozen=True)
            class Cls: pass
            """,
        ),
        Invalid(
            """
            import dataclasses
            @dataclasses.dataclass(other_kwarg=False)
            class Cls: pass
            """,
            line=2,
            expected_replacement="""
            import dataclasses
            @dataclasses.dataclass(other_kwarg=False, frozen=True)
            class Cls: pass
            """,
        ),
        Invalid(
            """
            from dataclasses import dataclass as dc
            @dc
            class Cls: pass
            """,
            line=2,
            expected_replacement="""
            from dataclasses import dataclass as dc
            @dc(frozen=True)
            class Cls: pass
            """,
        ),
        Invalid(
            """
            from dataclasses import dataclass as dc
            @dc()
            class Cls: pass
            """,
            line=2,
            expected_replacement="""
            from dataclasses import dataclass as dc
            @dc(frozen=True)
            class Cls: pass
            """,
        ),
        Invalid(
            """
            from dataclasses import dataclass as dc
            @dc(other_kwarg=False)
            class Cls: pass
            """,
            line=2,
            expected_replacement="""
            from dataclasses import dataclass as dc
            @dc(other_kwarg=False, frozen=True)
            class Cls: pass
            """,
        ),
        Invalid(
            """
            import dataclasses as dc
            @dc.dataclass
            class Cls: pass
            """,
            line=2,
            expected_replacement="""
            import dataclasses as dc
            @dc.dataclass(frozen=True)
            class Cls: pass
            """,
        ),
        Invalid(
            """
            import dataclasses as dc
            @dc.dataclass()
            class Cls: pass
            """,
            line=2,
            expected_replacement="""
            import dataclasses as dc
            @dc.dataclass(frozen=True)
            class Cls: pass
            """,
        ),
        Invalid(
            """
            import dataclasses as dc
            @dc.dataclass(other_kwarg=False)
            class Cls: pass
            """,
            line=2,
            expected_replacement="""
            import dataclasses as dc
            @dc.dataclass(other_kwarg=False, frozen=True)
            class Cls: pass
            """,
        ),
    ]

    def visit_ClassDef(self, node: cst.ClassDef) -> None:
        for d in node.decorators:
            decorator = d.decorator
            if QualifiedNameProvider.has_name(
                self,
                decorator,
                QualifiedName(
                    name="dataclasses.dataclass", source=QualifiedNameSource.IMPORT
                ),
            ):
                if isinstance(decorator, cst.Call):
                    func = decorator.func
                    args = decorator.args
                else:  # decorator is either cst.Name or cst.Attribute
                    args = ()
                    func = decorator

                # pyre-fixme[6]: Expected `Union[cst._maybe_sentinel.MaybeSentinel,
                #  cst._nodes.base.CSTNode, cst._removal_sentinel.RemovalSentinel]` for
                #  1st param but got `Optional[cst._nodes.expression.Name]`.
                if not any(m.matches(arg.keyword, m.Name("frozen")) for arg in args):
                    new_decorator = cst.Call(
                        func=func,
                        args=list(args)
                        + [
                            cst.Arg(
                                keyword=cst.Name("frozen"),
                                value=cst.Name("True"),
                                equal=cst.AssignEqual(
                                    whitespace_before=SimpleWhitespace(value=""),
                                    whitespace_after=SimpleWhitespace(value=""),
                                ),
                            )
                        ],
                    )
                    self.report(d, replacement=d.with_changes(decorator=new_decorator))



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import tokenize
from io import BytesIO
from typing import Mapping

from libcst.testing.utils import data_provider, UnitTest

from fixit.common.line_mapping import LineMappingInfo
from fixit.common.utils import dedent_with_lstrip


class LineMappingInfoTest(UnitTest):
    @data_provider(
        {
            "simple": {
                "code": dedent_with_lstrip(
                    """
                    def fn():
                        ...
                    """
                ),
                "physical_to_logical": {1: 1, 2: 2, 3: 3},
                "next_non_empty_logical_line": {1: 1, 2: 2, 3: 3},
            },
            "comments": {
                "code": dedent_with_lstrip(
                    """
                    # comment with
                    # multiple
                    # lines
                    def fn():
                        # comment
                        ...
                    """
                ),
                "physical_to_logical": {1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6, 7: 7},
                "next_non_empty_logical_line": {
                    1: 4,
                    2: 4,
                    3: 4,
                    4: 4,
                    5: 6,
                    6: 6,
                    7: 7,
                },
            },
            "blank_lines": {
                "code": dedent_with_lstrip(
                    """


                    def fn():


                        ...
                    """
                ),
                "physical_to_logical": {1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6, 7: 7},
                "next_non_empty_logical_line": {
                    1: 3,
                    2: 3,
                    3: 3,
                    4: 6,
                    5: 6,
                    6: 6,
                    7: 7,
                },
            },
            "line_continuation": {
                "code": dedent_with_lstrip(
                    """
                    value = "abc"
                    value = \\
                        "abcd" + \\
                        "efgh" + \\
                        "ijkl" + \\
                        "mnop"
                    """
                ),
                "physical_to_logical": {1: 1, 2: 2, 3: 2, 4: 2, 5: 2, 6: 2, 7: 7},
                "next_non_empty_logical_line": {
                    1: 1,
                    2: 2,
                    3: 7,
                    4: 7,
                    5: 7,
                    6: 7,
                    7: 7,
                },
            },
            "multiline_string": {
                "code": dedent_with_lstrip(
                    """
                    value = "abc"
                    value = '''
                        abcd
                        efgh
                        ijkl
                        mnop
                    '''
                    """
                ),
                "physical_to_logical": {1: 1, 2: 2, 3: 2, 4: 2, 5: 2, 6: 2, 7: 2, 8: 8},
                "next_non_empty_logical_line": {
                    1: 1,
                    2: 2,
                    3: 8,
                    4: 8,
                    5: 8,
                    6: 8,
                    7: 8,
                    8: 8,
                },
            },
        }
    )
    def test_line_mapping(
        self,
        *,
        code: str,
        physical_to_logical: Mapping[int, int],
        next_non_empty_logical_line: Mapping[int, int],
    ) -> None:
        tokens = tokenize.tokenize(BytesIO(code.encode("utf-8")).readline)
        result = LineMappingInfo.compute(tokens=tokens)
        self.assertEqual(dict(result.physical_to_logical), dict(physical_to_logical))
        for input, expected in next_non_empty_logical_line.items():
            self.assertEqual(result.get_next_non_empty_logical_line(input), expected)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from collections import defaultdict
from itertools import chain
from pathlib import Path
from subprocess import TimeoutExpired
from typing import Dict, Mapping
from unittest.mock import call, MagicMock, patch

from libcst.metadata import TypeInferenceProvider
from libcst.metadata.base_provider import ProviderT
from libcst.testing.utils import UnitTest

from fixit.common.full_repo_metadata import FullRepoMetadataConfig, get_repo_caches


class FullRepoMetadataTest(UnitTest):
    DUMMY_PATH = "fake/path.py"

    @patch("libcst.metadata.TypeInferenceProvider.gen_cache")
    def test_get_repo_caches_empty_paths(self, gen_cache: MagicMock) -> None:
        repo_caches: Mapping[str, Dict[ProviderT, object]] = get_repo_caches(
            [], FullRepoMetadataConfig({TypeInferenceProvider}, 1)
        )

        # We expect the call to resolve_cache to be bypassed.
        gen_cache.assert_not_called()
        self.assertEqual(repo_caches, {})

    @patch("libcst.metadata.TypeInferenceProvider.gen_cache")
    def test_get_repo_caches_single_path(self, gen_cache: MagicMock) -> None:
        gen_cache.return_value = {self.DUMMY_PATH: {}}
        repo_caches = get_repo_caches(
            (self.DUMMY_PATH,), FullRepoMetadataConfig({TypeInferenceProvider}, 1)
        )

        gen_cache.assert_called_with(Path(""), [self.DUMMY_PATH], 1)
        self.assertEqual(repo_caches, {self.DUMMY_PATH: {TypeInferenceProvider: {}}})

    @patch("libcst.metadata.TypeInferenceProvider.gen_cache")
    def test_get_repo_caches_swallows_exception(self, gen_cache: MagicMock) -> None:
        gen_cache.side_effect = Exception()
        repo_caches = get_repo_caches(
            (self.DUMMY_PATH,), FullRepoMetadataConfig({TypeInferenceProvider}, 1)
        )

        gen_cache.assert_called_with(Path(""), [self.DUMMY_PATH], 1)
        # Assert the placeholder cache is returned in the event of an error.
        self.assertEqual(
            repo_caches, {self.DUMMY_PATH: {TypeInferenceProvider: {"types": []}}}
        )

    @patch("libcst.metadata.TypeInferenceProvider.gen_cache")
    def test_get_repo_caches_honors_batch_size(self, gen_cache: MagicMock) -> None:
        paths = [f"{idx}_" + p for idx, p in enumerate([self.DUMMY_PATH] * 5)]
        gen_cache.return_value = {}

        get_repo_caches(
            paths, FullRepoMetadataConfig({TypeInferenceProvider}, 1, batch_size=2)
        )

        # A call to `gen_cache.__bool__()` will accompany every regular call to gen_cache() due to
        # the implementation of cache resolving in `libcst.metadata.FullRepoManager`.
        calls = [
            call(Path(""), paths[:2], 1),
            call(Path(""), paths[2:4], 1),
            call(Path(""), [paths[4]], 1),
        ]
        all_calls = chain.from_iterable([(call.__bool__(), c) for c in calls])

        # pyre-fixme[6]: Expected `Sequence[unittest.mock._Call]` for 1st param but
        #  got `Iterator[typing.Any]`.
        gen_cache.assert_has_calls(all_calls)

    @patch("libcst.metadata.TypeInferenceProvider.gen_cache")
    def test_get_repo_caches_with_logger(self, gen_cache: MagicMock) -> None:
        gen_cache.side_effect = TimeoutExpired("pyre query ...", 1)
        logger = logging.getLogger("Test")

        class CustomHandler(logging.Handler):
            errors = defaultdict(list)

            def emit(self, record: logging.LogRecord) -> None:
                exception_type = record.exc_info[0]
                self.errors[exception_type] += record.__dict__["paths"]

        hdlr = CustomHandler()
        logger.addHandler(hdlr)
        get_repo_caches(
            [self.DUMMY_PATH],
            FullRepoMetadataConfig(
                {TypeInferenceProvider}, 1, batch_size=2, logger=logger
            ),
        )

        self.assertEqual(CustomHandler.errors, {TimeoutExpired: [self.DUMMY_PATH]})



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from pathlib import Path

from jsonschema.exceptions import ValidationError
from libcst.testing.utils import UnitTest

from fixit.common.config import get_validated_settings


TEST_CONFIG = {
    "formatter": ["black", "-", "--no-diff"],
    "packages": ["python.fixit.rules"],
    "block_list_rules": ["Flake8PseudoLintRule"],
    "fixture_dir": f"{os.getcwd()}",
    "repo_root": f"{os.getcwd()}",
    "use_noqa": False,
    "rule_config": {
        "UnusedImportsRule": {
            "ignored_unused_modules": [
                "__future__",
                "__static__",
                "__static__.compiler_flags",
                "__strict__",
            ]
        },
    },
}


class TestConfig(UnitTest):
    def test_validated_settings_with_bad_types(self) -> None:
        bad_config = {"block_list_rules": False}
        with self.assertRaisesRegex(ValidationError, "False is not of type 'array'"):
            get_validated_settings(bad_config, Path("."))

    def test_validated_settings_with_correct_types(self) -> None:
        config = {"block_list_rules": ["FakeRule"]}
        settings = get_validated_settings(config, Path("."))
        self.assertEqual(
            {"block_list_rules": ["FakeRule"], "fixture_dir": ".", "repo_root": "."},
            settings,
        )

    def test_validated_settings_all_keys(self) -> None:
        self.maxDiff = None
        config = {
            "formatter": ["black", "-", "--no-diff"],
            "packages": ["python.fixit.rules"],
            "block_list_rules": ["Flake8PseudoLintRule"],
            "fixture_dir": f"{os.getcwd()}",
            "repo_root": f"{os.getcwd()}",
            "rule_config": {
                "UnusedImportsRule": {
                    "ignored_unused_modules": [
                        "__future__",
                        "__static__",
                        "__static__.compiler_flags",
                        "__strict__",
                    ]
                },
            },
        }
        settings = get_validated_settings(config, Path("."))
        self.assertEqual(
            config,
            settings,
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import importlib
from enum import Enum
from pathlib import Path
from typing import List

from fixit.common.base import LintRuleT
from fixit.common.config import get_lint_config, get_rules_from_config
from fixit.common.utils import (
    find_and_import_rule,
    import_distinct_rules_from_package,
    LintRuleNotFoundError,
)


class FixtureDirNotFoundError(Exception):
    pass


def import_rule(rule_name: str) -> LintRuleT:
    # Using the rule_name or full dotted name, attempt to import the rule.
    rule_module_path, _, rule_class_name = rule_name.rpartition(".")
    if rule_module_path:
        # If user provided a dotted path, we assume it's valid and import the rule directly.
        imported_rule = getattr(
            importlib.import_module(rule_module_path),
            rule_class_name,
        )
        return imported_rule
    # Otherwise, only a class name was provided, so try to find the rule by searching each package specified in the config.
    return find_and_import_rule(rule_class_name, get_lint_config().packages)


def get_pyre_fixture_dir_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "--fixture-dir",
        type=(lambda p: Path(p).resolve(strict=True)),
        help=("Main fixture file directory for integration testing."),
        default=get_lint_config().fixture_dir,
    )
    return parser


def get_rules_package_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "--rules-package",
        help=("Full dotted path of a package containing lint rules."),
        default="fixit.rules",
    )
    return parser


def get_rule_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "rule",
        type=import_rule,
        help=(
            "The name of your lint rule class or the full dotted path to your lint rule class. "
            + "(e.g. `NoAssertEqualsRule` or `fixit.rules.no_assert_equals.NoAssertEqualsRule`)"
        ),
    )
    return parser


class RuleAction(argparse.Action):
    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: List[str],
        option_string: str,
    ) -> None:
        all_rules = set()
        for rule_or_package in values:
            try:
                # Try to treat as a package first.
                all_rules.update(import_distinct_rules_from_package(rule_or_package))
            except ModuleNotFoundError:
                try:
                    all_rules.add(import_rule(rule_or_package))
                except LintRuleNotFoundError:
                    raise ValueError(
                        f"Unable to import rule or package named {rule_or_package}"
                    )
        setattr(namespace, self.dest, all_rules)


def get_rules_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "--rules",
        nargs="*",
        help=(
            "The names of lint rule classes to run, or packages containing lint rules, separated by a space. "
            + "(e.g `--rules NoAssertEqualsRule NoUnnecessaryListComprehensionRule my.custom.package`)"
        ),
        action=RuleAction,
        dest="rules",
        default=get_rules_from_config(),
    )
    return parser


def enforce_relative_to_repo_root(_path: str) -> Path:
    repo_root = get_lint_config().repo_root
    try:
        path = Path(_path).resolve(strict=True)
        path.relative_to(repo_root)
        return path
    except ValueError:
        raise argparse.ArgumentTypeError(
            f"Invalid value {_path}.\n"
            + f"Paths must be relative to the repo root `{repo_root}`"
            + " from the `repo_root` setting in the `.fixit.config.yaml` file."
        )


def get_paths_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "paths",
        nargs="*",
        type=enforce_relative_to_repo_root,
        default=(Path(get_lint_config().repo_root),),
        help=(
            "The name of a directory (e.g. media) or file (e.g. media/views.py) "
            + "relative to the `repo_root` specified in the `fixit.config.yaml` file"
            + " on which to run this script. "
            + "If not specified, the `repo_root` value is used."
        ),
    )
    return parser


def get_use_ignore_comments_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "--use-ignore-comments",
        action="store_true",
        default=True,
        help="Obey `# noqa`, `# lint-fixme` and `# lint-ignore` comments.",
    )
    return parser


def get_skip_ignore_comments_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "--skip-ignore-comments",
        dest="use_ignore_comments",
        action="store_false",
        help="Ignore `# noqa`, `# lint-fixme` and `# lint-ignore` comments.",
    )
    return parser


def get_skip_ignore_byte_marker_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "--skip-ignore-byte-markers",
        dest="use_ignore_byte_markers",
        action="store_false",
        help=f"Ignore `@gen{''}erated` and `@no{''}lint` markers in files.",
    )
    return parser


def get_metadata_cache_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "--cache-timeout",
        type=int,
        help="Timeout (seconds) for metadata cache fetching. Default is 2 seconds.",
        default=2,
    )
    return parser


def get_compact_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "--compact",
        action="store_true",
        help=(
            "Use a compact output that omits the message. This should be easier for "
            + "other scripts to parse."
        ),
    )
    return parser


def get_skip_autoformatter_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "--skip-autoformatter",
        action="store_true",
        help=(
            "Skips the autoformatter (e.g. Black) that's normally run after changes "
            + "are applied."
        ),
    )
    return parser


class LintWorkers(Enum):
    # Spawn (up to) one worker process per CPU core
    CPU_COUNT = "cpu_count"
    # Disable the process pool, and compute results in the current thread and process.
    #
    # This can be useful for debugging, where the process pool may break tracebacks,
    # debuggers, or profilers.
    USE_CURRENT_THREAD = "use_current_thread"


def get_multiprocessing_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "--no-multi",
        dest="workers",
        action="store_const",
        const=LintWorkers.USE_CURRENT_THREAD,
        default=LintWorkers.CPU_COUNT,
        help="Run the lint rule with multiprocessing disabled.",
    )
    return parser



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import ast
import pickle
from pathlib import Path

import libcst as cst
from libcst.testing.utils import data_provider, UnitTest

from fixit.common.report import AstLintRuleReport, BaseLintRuleReport, CstLintRuleReport


class LintRuleReportTest(UnitTest):
    @data_provider(
        {
            "AstLintRuleReport": [
                AstLintRuleReport(
                    file_path=Path("fake/path.py"),
                    node=ast.parse(""),
                    code="SomeFakeRule",
                    message="some message",
                    line=1,
                    column=1,
                )
            ],
            "CstLintRuleReport": [
                CstLintRuleReport(
                    file_path=Path("fake/path.py"),
                    node=cst.parse_statement("pass\n"),
                    code="SomeFakeRule",
                    message="some message",
                    line=1,
                    column=1,
                    module=cst.MetadataWrapper(cst.parse_module(b"pass\n")),
                    module_bytes=b"pass\n",
                )
            ],
        }
    )
    def test_is_not_pickleable(self, report: BaseLintRuleReport) -> None:
        with self.assertRaises(pickle.PicklingError):
            pickle.dumps(report)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Iterable

from fixit.common.flake8_compat import (
    Flake8LintRuleReport,
    get_cached_application_instance,
)
from fixit.common.pseudo_rule import PseudoLintRule


class Flake8PseudoLintRule(PseudoLintRule):
    """
    Executes all of flake8 under our lint engine as if it was a typical lint rule. This
    uses a bunch of internal flake8 APIs, so it's fragile and could break between
    releases of flake8.
    """

    def lint_file(self) -> Iterable[Flake8LintRuleReport]:
        app = get_cached_application_instance()
        app.reset(self.context)
        # TODO: use self.context.source, self.context.tokens, and self.context.ast
        app.run_checks([str(self.context.file_path)])
        # when we call report, it'll store the results to app.accumulator
        app.report()

        result = []
        for violation in app.accumulator:
            result.append(
                Flake8LintRuleReport(
                    file_path=self.context.file_path,
                    code=violation.code,
                    message=violation.text,
                    line=violation.line_number,
                    column=violation.column_number,
                )
            )

        return result



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from dataclasses import dataclass
from typing import Union

import libcst as cst
from libcst.metadata import (
    CodePosition,
    ExperimentalReentrantCodegenProvider,
    MetadataWrapper,
    ParentNodeProvider,
    WhitespaceInclusivePositionProvider,
)


@dataclass(frozen=True)
class LintPatch:
    """
    Represents the diff generated by an auto-fixer for arc lint.
    """

    # The point in source where the replacement was made. Typically this is the position
    # of the replaced node. Available as both a byte offset and a position.
    start_offset: int
    start_position: CodePosition
    # diff text
    original_diff_str: str
    patched_diff_str: str

    @staticmethod
    def get(
        wrapper: MetadataWrapper,
        original_node: cst.CSTNode,
        replacement_node: Union[cst.CSTNode, cst.FlattenSentinel, cst.RemovalSentinel],
    ) -> "LintPatch":
        # Batch the execution of these position providers
        wrapper.resolve_many(
            [
                ParentNodeProvider,
                ExperimentalReentrantCodegenProvider,
                WhitespaceInclusivePositionProvider,
            ]
        )

        # Use the resolve() API to fetch the data, because it's typed better than
        # resolve_many() is.
        parents = wrapper.resolve(ParentNodeProvider)
        positions = wrapper.resolve(WhitespaceInclusivePositionProvider)
        codegen_partials = wrapper.resolve(ExperimentalReentrantCodegenProvider)

        if isinstance(original_node, cst.Module) and isinstance(
            replacement_node, cst.RemovalSentinel
        ):
            raise Exception("Removing the entire module is not possible")

        # The reentrant codegen provider can only rewrite entire statements at a time,
        # so we need to inspect our parents until find a statement or the module
        possible_statement = original_node
        if isinstance(replacement_node, cst.RemovalSentinel):
            # reentrant codegen doesn't support RemovalSentinel, so use the parent instead
            possible_statement = parents[possible_statement]
        while True:
            if possible_statement in codegen_partials:
                partial = codegen_partials[possible_statement]
                patched_statement = cst.ensure_type(
                    _replace_or_remove(
                        possible_statement, original_node, replacement_node
                    ),
                    cst.BaseStatement,
                )
                original_str = partial.get_original_statement_code()
                patched_str = partial.get_modified_statement_code(patched_statement)
                return LintPatch(
                    partial.start_offset,
                    positions[possible_statement].start,
                    original_str,
                    patched_str,
                )
            elif possible_statement in parents:
                possible_statement = parents[possible_statement]
            else:
                # There's no more parents, so we have to fall back to replacing the whole
                # module.
                original_str = wrapper.module.code
                patched_module = cst.ensure_type(
                    _replace_or_remove(wrapper.module, original_node, replacement_node),
                    cst.Module,
                )
                patched_str = patched_module.code
                return LintPatch(0, CodePosition(1, 0), original_str, patched_str)

    def apply(self, original_module_str: str) -> str:
        return "".join(
            [
                original_module_str[: self.start_offset],
                self.patched_diff_str,
                original_module_str[self.start_offset + len(self.original_diff_str) :],
            ]
        )

    def minimize(self) -> "LintPatch":
        """
        Attempts to reduce the size of the patch. If you're going to apply multiple
        patches at once (without re-running the linter in between each patch), like
        `arc lint` does, you should attempt to minimize your patches to reduce the
        chance of overlapping changes.
        """
        return self._minimize_tail()._minimize_head()

    def _minimize_head(self) -> "LintPatch":
        """
        Strips any characters off of the head of the patch that are unchanged, reducing
        the size of the patch.
        """
        # line is 1-indexed, column is 0-indexed
        start_line, start_column = self.start_position.line, self.start_position.column

        matching_head_len = 0
        was_cr = False  # track \r\n newlines as a single newline
        for (a, b) in zip(self.original_diff_str, self.patched_diff_str):
            if a != b:
                break
            elif a == "\n":
                if not was_cr:
                    start_line += 1
                start_column = 0
                was_cr = False
            elif a == "\r":
                start_line += 1
                start_column = 0
                was_cr = True
            else:
                start_column += 1
                was_cr = False
            matching_head_len += 1

        return LintPatch(
            self.start_offset + matching_head_len,
            CodePosition(start_line, start_column),
            self.original_diff_str[matching_head_len:],
            self.patched_diff_str[matching_head_len:],
        )

    def _minimize_tail(self) -> "LintPatch":
        """
        Strips any characters off of the tail of the patch that are unchanged, reducing
        the size of the patch.
        """
        # this is easier than _minimize_head because we don't have to track line/column
        # numbers.
        matching_tail_len = 0
        for (a, b) in zip(
            reversed(self.original_diff_str), reversed(self.patched_diff_str)
        ):
            if a != b:
                break
            matching_tail_len += 1

        return LintPatch(
            self.start_offset,
            self.start_position,
            # We can't use `arr[:-offset]`` because when `offset` is `0`, we'd end up
            # with an empty string.
            self.original_diff_str[: len(self.original_diff_str) - matching_tail_len],
            self.patched_diff_str[: len(self.patched_diff_str) - matching_tail_len],
        )


def _replace_or_remove(
    parent: cst.CSTNode,
    original_node: cst.CSTNode,
    replacement_node: Union[cst.CSTNode, cst.FlattenSentinel, cst.RemovalSentinel],
) -> cst.CSTNode:
    if isinstance(replacement_node, (cst.RemovalSentinel, cst.FlattenSentinel)):
        return cst.ensure_type(parent.deep_remove(original_node), cst.CSTNode)
    else:
        return parent.deep_replace(original_node, replacement_node)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Sequence

import libcst as cst
import libcst.matchers as m

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


class NoAssertTrueForComparisonsRule(CstLintRule):
    """
    Finds incorrect use of ``assertTrue`` when the intention is to compare two values.
    These calls are replaced with ``assertEqual``.
    Comparisons with True, False and None are replaced with one-argument
    calls to ``assertTrue``, ``assertFalse`` and ``assertIsNone``.
    """

    MESSAGE: str = (
        '"assertTrue" does not compare its arguments, use "assertEqual" or other '
        + "appropriate functions."
    )

    VALID = [
        Valid("self.assertTrue(a == b)"),
        Valid('self.assertTrue(data.is_valid(), "is_valid() method")'),
        Valid("self.assertTrue(validate(len(obj.getName(type=SHORT))))"),
        Valid("self.assertTrue(condition, message_string)"),
    ]

    INVALID = [
        Invalid("self.assertTrue(a, 3)", expected_replacement="self.assertEqual(a, 3)"),
        Invalid(
            "self.assertTrue(hash(s[:4]), 0x1234)",
            expected_replacement="self.assertEqual(hash(s[:4]), 0x1234)",
        ),
        Invalid(
            "self.assertTrue(list, [1, 3])",
            expected_replacement="self.assertEqual(list, [1, 3])",
        ),
        Invalid(
            "self.assertTrue(optional, None)",
            expected_replacement="self.assertIsNone(optional)",
        ),
        Invalid(
            "self.assertTrue(b == a, True)",
            expected_replacement="self.assertTrue(b == a)",
        ),
        Invalid(
            "self.assertTrue(b == a, False)",
            expected_replacement="self.assertFalse(b == a)",
        ),
    ]

    def visit_Call(self, node: cst.Call) -> None:
        result = m.extract(
            node,
            m.Call(
                func=m.Attribute(value=m.Name("self"), attr=m.Name("assertTrue")),
                args=[
                    m.DoNotCare(),
                    m.Arg(
                        value=m.SaveMatchedNode(
                            m.OneOf(
                                m.Integer(),
                                m.Float(),
                                m.Imaginary(),
                                m.Tuple(),
                                m.List(),
                                m.Set(),
                                m.Dict(),
                                m.Name("None"),
                                m.Name("True"),
                                m.Name("False"),
                            ),
                            "second",
                        )
                    ),
                ],
            ),
        )

        if result:
            second_arg = result["second"]
            if isinstance(second_arg, Sequence):
                second_arg = second_arg[0]

            if m.matches(second_arg, m.Name("True")):
                new_call = node.with_changes(
                    args=[node.args[0].with_changes(comma=cst.MaybeSentinel.DEFAULT)],
                )
            elif m.matches(second_arg, m.Name("None")):
                new_call = node.with_changes(
                    func=node.func.with_deep_changes(
                        old_node=cst.ensure_type(node.func, cst.Attribute).attr,
                        value="assertIsNone",
                    ),
                    args=[node.args[0].with_changes(comma=cst.MaybeSentinel.DEFAULT)],
                )
            elif m.matches(second_arg, m.Name("False")):
                new_call = node.with_changes(
                    func=node.func.with_deep_changes(
                        old_node=cst.ensure_type(node.func, cst.Attribute).attr,
                        value="assertFalse",
                    ),
                    args=[node.args[0].with_changes(comma=cst.MaybeSentinel.DEFAULT)],
                )
            else:
                new_call = node.with_deep_changes(
                    old_node=cst.ensure_type(node.func, cst.Attribute).attr,
                    value="assertEqual",
                )

            self.report(node, replacement=new_call)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import tokenize
from io import BytesIO
from pathlib import Path
from typing import Container, Iterable, Tuple

import libcst as cst
from libcst.testing.utils import data_provider, UnitTest

from fixit.common.comments import CommentInfo
from fixit.common.ignores import IgnoreInfo
from fixit.common.line_mapping import LineMappingInfo
from fixit.common.report import CstLintRuleReport
from fixit.common.utils import dedent_with_lstrip


class IgnoreInfoTest(UnitTest):
    @data_provider(
        {
            # A noqa comment can be used without a specified code, which will ignore all
            # lint errors on that line. This is a bad practice, but we have to support
            # it for compatibility with existing code. (until we can codemod it away)
            "noqa_all_legacy": {
                "source": dedent_with_lstrip(
                    """
                    fn1()
                    fn2()  # noqa
                    fn3()
                    """
                ),
                "ignored_code": "IgnoredRule",
                "ignored_lines": [2],
            },
            # When a noqa comment is specified with codes, it should only ignore the
            # specified codes.
            "noqa_with_code": {
                "source": dedent_with_lstrip(
                    """
                    fn1()  # noqa: IgnoredRule
                    fn2()  # noqa: IgnoredRule: Message
                    fn3()  # noqa: IgnoredRule, Ignored2Rule: Message
                    fn4()  # noqa: Ignored1Rule
                    fn5()  # noqa: Ignored1Rule, Ignored2Rule
                    fn6()  # noqa: Ignored1Rule, Ignored2Rule: Message
                    """
                ),
                "ignored_code": "Ignored1Rule",
                "ignored_lines": [4, 5, 6],
            },
            "noqa_multiline": {
                "source": dedent_with_lstrip(
                    """
                    fn1(line, \\
                    continuation)  # noqa: IgnoredRule

                    fn2()

                    fn3('''
                        multiline
                        string
                    ''')  # noqa: IgnoredRule
                    """
                ),
                "ignored_code": "IgnoredRule",
                "ignored_lines": [1, 2, 6, 7, 8, 9],
            },
            "noqa_file": {
                "source": dedent_with_lstrip(
                    """
                    # noqa-file: IgnoredRule: Some reason
                    fn1()
                    """
                ),
                "ignored_code": "IgnoredRule",
                "ignored_lines": [1, 2, 3],
            },
            "noqa_file_multiple_codes": {
                "source": dedent_with_lstrip(
                    """
                    # noqa-file: IgnoredRule, Ignored1Rule, Ignored2Rule: Some reason
                    fn1()
                    """
                ),
                "ignored_code": "Ignored1Rule",
                "ignored_lines": [1, 2, 3],
            },
            "noqa_file_requires_code_and_reason": {
                "source": dedent_with_lstrip(
                    """
                    # noqa-file
                    # noqa-file: IgnoredRule
                    # Neither of these noqa-files should work because they're incomplete
                    fn1()
                    """
                ),
                "ignored_code": "IgnoredRule",
                "ignored_lines": [],
            },
            "backwards_compatibility_classname": {
                "source": dedent_with_lstrip(
                    """
                    fn1() # noqa: IG00, IgnoredRule
                    """
                ),
                "ignored_code": "IgnoredRule",
                "ignored_lines": [1],
            },
            "backwards_compatibility_oldcode": {
                "source": dedent_with_lstrip(
                    """
                    fn1() # noqa: IG00, IgnoredRule
                    """
                ),
                "ignored_code": "IG00",
                "ignored_lines": [1],
            },
            "lint_fixme": {
                "source": dedent_with_lstrip(
                    """
                    fn1()

                    # lint-fixme: IgnoredRule: Some short reason
                    fn2(  # this line should be ignored
                        "multiple",  # but these lines shouldn't
                        "arguments",
                    )

                    # lint-fixme: IgnoredRule: Some reason spanning
                    # lint: multiple lines because it's long.
                    fn3('''
                        multiline
                        string
                    ''')  # this function call is a single logical line

                    fn4()
                    """
                ),
                "ignored_code": "IgnoredRule",
                "ignored_lines": [3, 4, 9, 10, 11, 12, 13, 14],
            },
            "lint_ignore": {
                "source": dedent_with_lstrip(
                    """
                    fn1()

                    # lint-ignore: IgnoredRule: Some reason
                    fn2()

                    fn3()
                    """
                ),
                "ignored_code": "IgnoredRule",
                "ignored_lines": [3, 4],
            },
            # A lint-ignore can exist right before an EOF. That's fine. We should ignore
            # all the way to the EOF.
            "lint_ignore_eof": {
                "source": dedent_with_lstrip(
                    """
                    # lint-ignore: IgnoredRule
                    """
                ),
                "ignored_code": "IgnoredRule",
                "ignored_lines": [1, 2],
            },
        }
    )
    def test_ignored_lines(
        self, *, source: str, ignored_code: str, ignored_lines: Container[int]
    ) -> None:
        tokens = tuple(tokenize.tokenize(BytesIO(source.encode("utf-8")).readline))
        ignore_info = IgnoreInfo.compute(
            comment_info=CommentInfo.compute(tokens=tokens),
            line_mapping_info=LineMappingInfo.compute(tokens=tokens),
            use_noqa=True,
        )
        lines = range(1, tokens[-1].end[0] + 1)
        actual_ignored_lines = []
        for line in lines:
            ignored = ignore_info.should_ignore_report(
                CstLintRuleReport(
                    file_path=Path("fake/path.py"),
                    node=cst.EmptyLine(),
                    code=ignored_code,
                    message="message",
                    line=line,
                    column=0,
                    module=cst.MetadataWrapper(cst.parse_module(source)),
                    module_bytes=source.encode("utf-8"),
                )
            )
            if ignored:
                actual_ignored_lines.append(line)
        # pyre-fixme[6]: Expected `Iterable[Variable[_T]]` for 1st param but got
        #  `Container[int]`.
        self.assertEqual(actual_ignored_lines, list(ignored_lines))

    @data_provider(
        {
            "unused_noqa": {
                "source": "fn()  # noqa",
                "reports_on_lines": [],
                "unused_comments": [1],
            },
            "used_noqa": {
                "source": "fn()  # noqa",
                "reports_on_lines": [(1, "Ignored999Rule")],
                "unused_comments": [],
            },
            "unused_lint_ignore": {
                "source": "# lint-ignore: Ignored999Rule: Some reason\nfn()",
                "reports_on_lines": [],
                "unused_comments": [1],
            },
            "unused_lint_ignore_mutliple_lines": {
                "source": "# lint-ignore: Ignored999Rule: Some\n# lint: reason blah blah blah\nfn()",
                "reports_on_lines": [],
                "unused_comments": [1],
            },
            "used_lint_ignore": {
                "source": "# lint-ignore: Ignored999Rule: Some reason\nfn()",
                "reports_on_lines": [(2, "Ignored999Rule")],
                "unused_comments": [],
            },
            "used_lint_ignore_multiple_lines": {
                "source": "# lint-ignore: Ignored999Rule: Some\n# lint: reason blah blah blah\nfn()",
                "reports_on_lines": [(3, "Ignored999Rule")],
                "unused_comments": [],
            },
            "lint_ignore_is_used_before_noqa": {
                "source": "# lint-ignore: Ignored999Rule: Some reason\nfn()  # noqa",
                "reports_on_lines": [(2, "Ignored999Rule")],
                "unused_comments": [2],
            },
            "duplicate_lint_ignores": {
                "source": "# lint-ignore: Ignored999Rule: First\n# lint-ignore: Ignored999Rule: Second\nfn()",
                "reports_on_lines": [(3, "Ignored999Rule")],
                "unused_comments": [2],
            },
            "multiple_used_lint_ignores": {
                "source": dedent_with_lstrip(
                    """
                    # lint-ignore: Ignored999Rule: Some
                    # lint: reason blah blah blah
                    # lint-ignore: Ignored1000Rule: Some
                    # lint: other reason blah blah
                    fn()
                    """
                ),
                "reports_on_lines": [(5, "Ignored999Rule"), (5, "Ignored1000Rule")],
                "unused_comments": [],
            },
            "multiple_unused_lint_ignores": {
                "source": dedent_with_lstrip(
                    """
                    # lint-ignore: Ignored999Rule: Some
                    # lint: reason blah blah blah
                    # lint-ignore: Ignored1000Rule: Some
                    # lint: other reason blah blah
                    fn()
                    """
                ),
                "reports_on_lines": [],
                "unused_comments": [1, 3],
            },
            "some_unused_lint_ignores": {
                "source": dedent_with_lstrip(
                    """
                    # lint-ignore: Ignored999Rule: Some
                    # lint: reason blah blah blah
                    # lint-ignore: Ignored1000Rule: Some
                    # lint: other reason blah blah
                    fn()
                    """
                ),
                "reports_on_lines": [(5, "Ignored999Rule")],
                "unused_comments": [3],
            },
            "some_unused_lint_ignores_2": {
                "source": dedent_with_lstrip(
                    """
                    # lint-ignore: Ignored999Rule: Some
                    # lint: reason blah blah blah
                    # lint-ignore: Ignored1000Rule: Some
                    # lint: other reason blah blah
                    fn()
                    """
                ),
                "reports_on_lines": [(5, "Ignored1000Rule")],
                "unused_comments": [1],
            },
        }
    )
    def test_unused_comments(
        self,
        *,
        source: str,
        reports_on_lines: Iterable[Tuple[int, str]],
        unused_comments: Iterable[int],
    ) -> None:
        """
        Verify that we can correctly track which lint comments were used and which were
        unused.

        TODO: We don't track usage of global ignore comments, so we can't know if
        they're unused.
        """
        tokens = tuple(tokenize.tokenize(BytesIO(source.encode("utf-8")).readline))
        ignore_info = IgnoreInfo.compute(
            comment_info=CommentInfo.compute(tokens=tokens),
            line_mapping_info=LineMappingInfo.compute(tokens=tokens),
            use_noqa=True,
        )

        for line, code in reports_on_lines:
            ignore_info.should_ignore_report(
                CstLintRuleReport(
                    file_path=Path("fake/path.py"),
                    node=cst.EmptyLine(),
                    code=code,
                    message="message",
                    line=line,
                    column=0,
                    module=cst.MetadataWrapper(cst.parse_module(source)),
                    module_bytes=source.encode("utf-8"),
                )
            )

        self.assertEqual(
            sorted(
                [
                    min(tok.start[0] for tok in c.tokens)
                    for c in ignore_info.suppression_comments
                    if not c.used_by
                ]
            ),
            sorted(unused_comments),
        )

    def test_multiline_suppression(self) -> None:
        source = """
        # lint-ignore: SomeCode: some reason
        # lint: and some reason continued
        # lint: onto multiple lines.
        x = "Some ignored violation"
        """
        tokens = tuple(tokenize.tokenize(BytesIO(source.encode("utf-8")).readline))
        ignore_info = IgnoreInfo.compute(
            comment_info=CommentInfo.compute(tokens=tokens),
            line_mapping_info=LineMappingInfo.compute(tokens=tokens),
            use_noqa=False,
        )

        self.assertEqual(
            len(ignore_info.local_ignore_info.local_suppression_comments), 1
        )
        local_supp_comment = next(
            lsc for lsc in ignore_info.local_ignore_info.local_suppression_comments
        )
        self.assertEqual(
            local_supp_comment.reason,
            "some reason and some reason continued onto multiple lines.",
        )
        # Verify that all local suppression comment lines map to the same SuppressionComment instance.
        for (
            lines,
            supp_comments,
        ) in ignore_info.local_ignore_info.local_suppression_comments_by_line.items():
            self.assertEqual(len(supp_comments), 1)
            supp_comment = supp_comments[0]
            self.assertIs(supp_comment, local_supp_comment)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import libcst as cst
import libcst.matchers as m

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


UNNECESSARY_LIST_COMPREHENSION: str = (
    "Unnecessary list comprehension - {func} can take a generator, and is likely "
    + "to short-circuit, so constructing a list is probably wasteful."
)


class NoRedundantListComprehensionRule(CstLintRule):
    """
    A derivative of flake8-comprehensions's C407 rule.
    """

    VALID = [
        Valid("any(val for val in iterable)"),
        Valid("all(val for val in iterable)"),
        # C407 would complain about these, but we won't
        Valid("frozenset([val for val in iterable])"),
        Valid("max([val for val in iterable])"),
        Valid("min([val for val in iterable])"),
        Valid("sorted([val for val in iterable])"),
        Valid("sum([val for val in iterable])"),
        Valid("tuple([val for val in iterable])"),
    ]
    INVALID = [
        Invalid(
            "any([val for val in iterable])",
            expected_replacement="any(val for val in iterable)",
        ),
        Invalid(
            "all([val for val in iterable])",
            expected_replacement="all(val for val in iterable)",
        ),
    ]

    def visit_Call(self, node: cst.Call) -> None:
        # This set excludes frozenset, max, min, sorted, sum, and tuple, which C407 would warn
        # about, because none of those functions short-circuit.
        if m.matches(
            node,
            m.Call(
                func=m.Name("all") | m.Name("any"), args=[m.Arg(value=m.ListComp())]
            ),
        ):
            list_comp = cst.ensure_type(node.args[0].value, cst.ListComp)
            self.report(
                node,
                UNNECESSARY_LIST_COMPREHENSION.format(
                    func=cst.ensure_type(node.func, cst.Name).value
                ),
                replacement=node.deep_replace(
                    list_comp,
                    cst.GeneratorExp(
                        elt=list_comp.elt, for_in=list_comp.for_in, lpar=[], rpar=[]
                    ),
                ),
            )



# Copyright (c) Meta Platforms, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path

from .types import Config


def generate_config(path: Path) -> Config:
    """
    Given a file path, walk upwards looking for and applying cascading configs
    """
    path = path.resolve()

    return Config(
        path=path,
    )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import sys

from fixit.common.config import gen_config_file


if __name__ == "__main__":
    """
    Generate a `fixit.config.yaml` file with defaults in the current working directory.
    """
    gen_config_file()
    sys.exit(0)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from libcst.codemod.commands.noop import NOOPCommand
from libcst.testing.utils import UnitTest

from fixit.cli.args import get_rule_parser


class SomeFakeRule:
    pass


class LintRuleCLIArgsTest(UnitTest):
    def test_rule_parser(self) -> None:
        parser = get_rule_parser().parse_args(
            ["fixit.cli.tests.test_args.SomeFakeRule"]
        )
        self.assertEqual(parser.rule, SomeFakeRule)

    def test_rule_parser_external_module(self) -> None:
        # External modules work, as long as they are a dependency
        parser = get_rule_parser().parse_args(
            ["libcst.codemod.commands.noop.NOOPCommand"]
        )
        self.assertEqual(parser.rule, NOOPCommand)



# Copyright (c) Meta Platforms, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from unittest import TestCase

from click.testing import CliRunner

from fixit import __version__
from fixit.cli import main


class SmokeTest(TestCase):
    def setUp(self):
        self.runner = CliRunner(mix_stderr=False)

    def test_cli_version(self):
        result = self.runner.invoke(main, ["--version"])
        self.assertRegex(result.stdout, rf"fixit, version {__version__}")



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import cast, Dict, List, Optional, Union

import libcst as cst
import libcst.matchers as m
from libcst.metadata import TypeInferenceProvider

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid
from fixit.common.report import CstLintRuleReport


class UseIsNoneOnOptionalRule(CstLintRule):
    """
    Enforces explicit use of ``is None`` or ``is not None`` when checking whether an Optional has a value.
    Directly testing an object (e.g. ``if x``) implicitely tests for a truth value which returns ``True`` unless the
    object's ``__bool__()`` method returns False, its ``__len__()`` method returns '0', or it is one of the constants ``None`` or ``False``.
    (https://docs.python.org/3.8/library/stdtypes.html#truth-value-testing).
    """

    METADATA_DEPENDENCIES = (TypeInferenceProvider,)
    MESSAGE: str = (
        "When checking if an `Optional` has a value, avoid using it as a boolean since it implicitly checks the object's `__bool__()`, `__len__()` is not `0`, or the value is not `None`. "
        + "Instead, use `is None` or `is not None` to be explicit."
    )

    VALID: List[Valid] = [
        Valid(
            """
            from typing import Optional
            a: Optional[str]
            if a is not None:
                pass
            """,
        ),
        Valid(
            """
            a: bool
            if a:
                pass
            """,
        ),
    ]

    INVALID: List[Invalid] = [
        Invalid(
            code="""
            from typing import Optional

            a: Optional[str] = None
            if a:
                pass
            """,
            expected_replacement="""
            from typing import Optional

            a: Optional[str] = None
            if a is not None:
                pass
            """,
        ),
        Invalid(
            code="""
            from typing import Optional
            a: Optional[str] = None
            x: bool = False
            if x and a:
                ...
            """,
            expected_replacement="""
            from typing import Optional
            a: Optional[str] = None
            x: bool = False
            if x and a is not None:
                ...
            """,
        ),
        Invalid(
            code="""
            from typing import Optional
            a: Optional[str] = None
            x: bool = False
            if a and x:
                ...
            """,
            expected_replacement="""
            from typing import Optional
            a: Optional[str] = None
            x: bool = False
            if a is not None and x:
                ...
            """,
        ),
        Invalid(
            code="""
            from typing import Optional
            a: Optional[str] = None
            x: bool = not a
            """,
            expected_replacement="""
            from typing import Optional
            a: Optional[str] = None
            x: bool = a is None
            """,
        ),
        Invalid(
            code="""
            from typing import Optional
            a: Optional[str]
            x: bool
            if x or a: pass
            """,
            expected_replacement="""
            from typing import Optional
            a: Optional[str]
            x: bool
            if x or a is not None: pass
            """,
        ),
        Invalid(
            code="""
            from typing import Optional
            a: Optional[str]
            x: bool
            if x: pass
            elif a: pass
            """,
            expected_replacement="""
            from typing import Optional
            a: Optional[str]
            x: bool
            if x: pass
            elif a is not None: pass
            """,
        ),
        Invalid(
            code="""
            from typing import Optional
            a: Optional[str] = None
            b: Optional[str] = None
            if a: pass
            elif b: pass
            """,
            expected_replacement="""
            from typing import Optional
            a: Optional[str] = None
            b: Optional[str] = None
            if a is not None: pass
            elif b is not None: pass
            """,
        ),
    ]

    def leave_If(self, original_node: cst.If) -> None:
        changes: Dict[str, cst.CSTNode] = {}
        test_expression: cst.BaseExpression = original_node.test
        if m.matches(test_expression, m.Name()):
            # We are inside a simple check such as "if x".
            test_expression = cast(cst.Name, test_expression)
            if self._is_optional_type(test_expression):
                # We want to replace "if x" with "if x is not None".
                replacement_comparison: cst.Comparison = self._gen_comparison_to_none(
                    variable_name=test_expression.value, operator=cst.IsNot()
                )
                changes["test"] = replacement_comparison

        orelse = original_node.orelse
        if orelse is not None and m.matches(orelse, m.If()):
            # We want to catch this case upon leaving an `If` node so that we generate an `elif` statement correctly.
            # We check if the orelse node was reported, and if so, remove the report and generate a new report on
            # the current parent `If` node.
            new_reports = []
            orelse_report: Optional[CstLintRuleReport] = None
            for report in self.context.reports:
                if isinstance(report, CstLintRuleReport):
                    # Check whether the lint rule code matches this lint rule's code so we don't remove another
                    # lint rule's report.
                    if report.node is orelse and report.code == self.__class__.__name__:
                        orelse_report = report
                    else:
                        new_reports.append(report)
                else:
                    new_reports.append(report)
            if orelse_report is not None:
                self.context.reports = new_reports
                replacement_orelse = orelse_report.replacement_node
                changes["orelse"] = cst.ensure_type(replacement_orelse, cst.CSTNode)

        if changes:
            self.report(
                original_node, replacement=original_node.with_changes(**changes)
            )

    def visit_BooleanOperation(self, node: cst.BooleanOperation) -> None:
        left_expression: cst.BaseExpression = node.left
        right_expression: cst.BaseExpression = node.right
        if m.matches(node.left, m.Name()):
            # Eg: "x and y".
            left_expression = cast(cst.Name, left_expression)
            if self._is_optional_type(left_expression):
                replacement_comparison = self._gen_comparison_to_none(
                    variable_name=left_expression.value, operator=cst.IsNot()
                )
                self.report(
                    node, replacement=node.with_changes(left=replacement_comparison)
                )
        if m.matches(right_expression, m.Name()):
            # Eg: "x and y".
            right_expression = cast(cst.Name, right_expression)
            if self._is_optional_type(right_expression):
                replacement_comparison = self._gen_comparison_to_none(
                    variable_name=right_expression.value, operator=cst.IsNot()
                )
                self.report(
                    node, replacement=node.with_changes(right=replacement_comparison)
                )

    def visit_UnaryOperation(self, node: cst.UnaryOperation) -> None:
        if m.matches(node, m.UnaryOperation(operator=m.Not(), expression=m.Name())):
            # Eg: "not x".
            expression: cst.Name = cast(cst.Name, node.expression)
            if self._is_optional_type(expression):
                replacement_comparison = self._gen_comparison_to_none(
                    variable_name=expression.value, operator=cst.Is()
                )
                self.report(node, replacement=replacement_comparison)

    def _is_optional_type(self, node: cst.Name) -> bool:
        reported_type = self.get_metadata(TypeInferenceProvider, node, None)
        # We want to use `startswith()` here since the type data will take on the form 'typing.Optional[SomeType]'.
        if reported_type is not None and reported_type.startswith("typing.Optional"):
            return True
        return False

    def _gen_comparison_to_none(
        self, variable_name: str, operator: Union[cst.Is, cst.IsNot]
    ) -> cst.Comparison:
        return cst.Comparison(
            left=cst.Name(value=variable_name),
            comparisons=[
                cst.ComparisonTarget(
                    operator=operator, comparator=cst.Name(value="None")
                )
            ],
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from pathlib import Path

import libcst as cst
from libcst.testing.utils import UnitTest

from fixit.cli.formatter import format_warning, LintRuleReportFormatter
from fixit.common.report import BaseLintRuleReport, CstLintRuleReport


class _ExtendedLintRuleReportFormatter(LintRuleReportFormatter):
    def _format_position(self, report: BaseLintRuleReport) -> str:
        return "<overridden position>"

    def _format_details_raw(self, report: BaseLintRuleReport) -> str:
        return "<overridden details_raw>"


class LintRuleReportFormatterTest(UnitTest):
    def setUp(self) -> None:
        self.fake_filepath = Path("fake/path.py")
        self.report = CstLintRuleReport(
            file_path=self.fake_filepath,
            node=cst.parse_statement("pass\n"),
            code="SomeFakeRule",
            message=(
                "Some long message that should span multiple lines.\n"
                + "\n"
                + "Another paragraph with more information about the lint rule."
            ),
            line=1,
            column=1,
            module=cst.MetadataWrapper(cst.parse_module(b"pass\n")),
            module_bytes=b"pass\n",
        )

    def test_format_full(self) -> None:
        full_formatter = LintRuleReportFormatter(width=20, compact=False)
        self.assertEqual(
            full_formatter.format(self.report),
            (
                str(self.fake_filepath)
                + ":1:1\n"
                + "    SomeFakeRule:\n"
                + "    Some long\n"
                + "    message that\n"
                + "    should span\n"
                + "    multiple lines.\n"
                + "\n"
                + "    Another\n"
                + "    paragraph with\n"
                + "    more information\n"
                + "    about the lint\n"
                + "    rule."
            ),
        )

    def test_format_compact(self) -> None:
        compact_formatter = LintRuleReportFormatter(width=20, compact=True)
        self.assertEqual(
            compact_formatter.format(self.report),
            str(self.fake_filepath) + ":1:1",
        )

    def test_format_extended(self) -> None:
        extended_formatter = _ExtendedLintRuleReportFormatter(width=100, compact=False)
        self.assertEqual(
            extended_formatter.format(self.report),
            "<overridden position>\n    <overridden details_raw>",
        )


class FormatWarningTest(UnitTest):
    def test_format_warning(self) -> None:
        self.assertEqual(
            format_warning("Some long warning message that should be wrapped.", 20),
            (
                "Warning: Some long\n"
                + "         warning\n"
                + "         message\n"
                + "         that should\n"
                + "         be wrapped."
            ),
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import libcst as cst

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


class UseLintFixmeCommentRule(CstLintRule):
    """
    To silence a lint warning, use ``lint-fixme`` (when plans to fix the issue later) or ``lint-ignore``
    (when the lint warning is not valid) comments.
    The comment requires to be in a standalone comment line and follows the format ``lint-fixme: RULE_NAMES EXTRA_COMMENTS``.
    It suppresses the lint warning with the RULE_NAMES in the next line.
    RULE_NAMES can be one or more lint rule names separated by comma.
    ``noqa`` is deprecated and not supported because explicitly providing lint rule names to be suppressed
    in lint-fixme comment is preferred over implicit noqa comments. Implicit noqa suppression comments
    sometimes accidentally silence warnings unexpectedly.
    """

    MESSAGE: str = "noqa is deprecated. Use `lint-fixme` or `lint-ignore` instead."

    VALID = [
        Valid(
            """
            # lint-fixme: UseFstringRule
            "%s" % "hi"
            """
        ),
        Valid(
            """
            # lint-ignore: UsePlusForStringConcatRule
            'ab' 'cd'
            """
        ),
    ]
    INVALID = [
        Invalid("fn() # noqa"),
        Invalid(
            """
            (
             1,
             2,  # noqa
            )
            """
        ),
        Invalid(
            """
            class C:
                # noqa
                ...
            """
        ),
    ]

    def visit_Comment(self, node: cst.Comment) -> None:
        target = "# noqa"
        if node.value[: len(target)].lower() == target:
            self.report(node)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import libcst as cst
import libcst.matchers as m

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


class AvoidOrInExceptRule(CstLintRule):
    """
    Discourages use of ``or`` in except clauses. If an except clause needs to catch multiple exceptions,
    they must be expressed as a parenthesized tuple, for example:
    ``except (ValueError, TypeError)``
    (https://docs.python.org/3/tutorial/errors.html#handling-exceptions)

    When ``or`` is used, only the first operand exception type of the conditional statement will be caught.
    For example::

        In [1]: class Exc1(Exception):
            ...:     pass
            ...:

        In [2]: class Exc2(Exception):
            ...:     pass
            ...:

        In [3]: try:
            ...:     raise Exception()
            ...: except Exc1 or Exc2:
            ...:     print("caught!")
            ...:
        ---------------------------------------------------------------------------
        Exception                                 Traceback (most recent call last)
        <ipython-input-3-3340d66a006c> in <module>
            1 try:
        ----> 2     raise Exception()
            3 except Exc1 or Exc2:
            4     print("caught!")
            5

        Exception:

        In [4]: try:
            ...:     raise Exc1()
            ...: except Exc1 or Exc2:
            ...:     print("caught!")
            ...:
            caught!

        In [5]: try:
            ...:     raise Exc2()
            ...: except Exc1 or Exc2:
            ...:     print("caught!")
            ...:
        ---------------------------------------------------------------------------
        Exc2                                      Traceback (most recent call last)
        <ipython-input-5-5d29c1589cc0> in <module>
            1 try:
        ----> 2     raise Exc2()
            3 except Exc1 or Exc2:
            4     print("caught!")
            5

        Exc2:
    """

    MESSAGE: str = (
        "Avoid using 'or' in an except block. For example:"
        + "'except ValueError or TypeError' only catches 'ValueError'. Instead, use "
        + "parentheses, 'except (ValueError, TypeError)'"
    )
    VALID = [
        Valid(
            """
            try:
                print()
            except (ValueError, TypeError) as err:
                pass
            """
        )
    ]

    INVALID = [
        Invalid(
            """
            try:
                print()
            except ValueError or TypeError:
                pass
            """,
        )
    ]

    def visit_Try(self, node: cst.Try) -> None:
        if m.matches(
            node,
            m.Try(handlers=[m.ExceptHandler(type=m.BooleanOperation(operator=m.Or()))]),
        ):
            self.report(node)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from dataclasses import dataclass
from enum import Enum
from fnmatch import fnmatch
from functools import lru_cache
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Sequence, Set

import libcst as cst
from libcst.helpers import get_full_name_for_node_or_raise
from libcst.metadata import GlobalScope, ScopeProvider

from fixit import (
    CstContext,
    CstLintRule,
    InvalidTestCase as Invalid,
    LintConfig,
    ValidTestCase as Valid,
)


TEST_REPO_ROOT: str = str(Path(__file__).parent.parent)


class RuleAction(Enum):
    ALLOW = "allow"  # allow everywhere
    DENY = "deny"  # deny everywhere
    ALLOW_GLOBAL = "allow_global"  # allow in global scope
    ALLOW_LOCAL = "allow_local"  # allow in local scopes
    DENY_GLOBAL = "deny_global"  # deny in global scope
    DENY_LOCAL = "deny_local"  # deny in local scopes


def _gen_testcase_config(dir_rules: Dict[str, object]) -> LintConfig:
    return LintConfig(
        repo_root=TEST_REPO_ROOT, rule_config={"ImportConstraintsRule": dir_rules}
    )


@dataclass(frozen=True)
class _ImportRule:
    pattern: str  # dot-separated module/package name prefix, or wildcard (*)
    action: RuleAction

    @staticmethod
    def from_config(rule: object) -> "_ImportRule":
        if not isinstance(rule, list) or len(rule) != 2:
            raise ValueError(
                f"Invalid entry `{rule}`.\n"
                + "Each rule under a directory must specify a module and an action."
                + ' E.g. \'["*", "deny"]\''
            )
        else:
            module, action = rule
            return _ImportRule(module, RuleAction(action))

    @property
    def is_wildcard(self) -> bool:
        return self.pattern == "*"

    def match(self, global_name: str) -> bool:
        if self.is_wildcard:
            return True

        split_pattern = self.pattern.split(".")
        return global_name.split(".")[: len(split_pattern)] == split_pattern


@dataclass(frozen=True)
class _ImportConfig:
    rules: Sequence[_ImportRule]
    ignore_tests: bool
    ignore_types: bool
    message: Optional[str]

    @staticmethod
    def from_config(settings_for_dir: Dict[object, object]) -> "_ImportConfig":
        rule_settings_for_dir = settings_for_dir.get("rules", [])

        if not isinstance(rule_settings_for_dir, list):
            raise ValueError(
                f"Invalid entry `{rule_settings_for_dir}`.\n"
                + "The `rules` setting must be a list of rules."
            )

        rules_for_file = []
        for rule in rule_settings_for_dir:
            rules_for_file.append(_ImportRule.from_config(rule))

        ignore_tests = settings_for_dir.get("ignore_tests", True)
        ignore_types = settings_for_dir.get("ignore_types", True)
        message = settings_for_dir.get("message")
        if not isinstance(ignore_tests, bool):
            raise ValueError("Setting `ignore_tests` value must be 'True' or 'False'.")
        if not isinstance(ignore_types, bool):
            raise ValueError("Setting `ignore_types` value must be 'True' or 'False'.")
        if message is not None and not isinstance(message, str):
            raise ValueError("Setting `message` value must be a string.")

        return _ImportConfig(
            rules_for_file, ignore_tests, ignore_types, message
        )._validate()

    def _validate(self) -> "_ImportConfig":
        if len(self.rules) == 0:
            raise ValueError("Must have at least one rule")
        if not self.rules[-1].is_wildcard:
            raise ValueError("The last rule must be a wildcard rule")
        if any(r.is_wildcard for r in self.rules[:-1]):
            raise ValueError("Only the last rule can be a wildcard rule")
        return self

    def match(self, global_name: str) -> "_ImportRule":
        for r in self.rules:
            if r.match(global_name):
                return r
        raise AssertionError(
            "No matching rule was found. The wildcard rule should make this impossible."
        )


@lru_cache(maxsize=None)
def _get_local_roots(repo_root: Path) -> Set[str]:
    return set(os.listdir(repo_root))


class ImportConstraintsRule(CstLintRule):
    """
    Rule to impose import constraints in certain directories to improve runtime performance.
    The directories specified in the ImportConstraintsRule setting in the ``.fixit.config.yaml`` file's
    ``rule_config`` section can impose import constraints for that directory and its children as follows::

        rule_config:
            ImportConstraintsRule:
                dir_under_repo_root:
                    rules: [
                        ["module_under_repo_root", "allow"],
                        ["another_module_under_repo_root, "deny"],
                        ["*", "deny"]
                    ]
                    ignore_tests: True
                    ignore_types: True
                    message: "'{imported}' cannot be imported from within '{current_file}'."

    Each rule under ``rules`` is evaluated in order from top to bottom and the last rule for each directory
    should be a wildcard rule. Rules can be ``"allow"``, ``"deny"``, ``"allow_global"``, ``"allow_local"``,
    ``"deny_global"`` or ``"deny_local"``.
    ``ignore_tests`` and `ignore_types` should carry boolean values and can be omitted. They are both set to
    `True` by default.
    If ``ignore_types`` is True, this rule will ignore imports inside ``if TYPE_CHECKING`` blocks since those
    imports do not have an affect on runtime performance.
    If ``ignore_tests`` is True, this rule will not lint any files found in a testing module.
    If ``message`` is passed, it must be a string containing a custom message. The string will be formatted
    passing ``imported`` and ``current_file`` variables to be used if needed, with the symbol being imported
    and the current file, respectively.
    """

    _config: Optional[_ImportConfig]
    _repo_root: Path
    _type_checking_stack: List[cst.If]
    _abs_file_path: Path

    METADATA_DEPENDENCIES = (ScopeProvider,)
    MESSAGE: str = (
        "According to the settings for this directory in the .fixit.config.yaml configuration file, "
        + "'{imported}' cannot be imported from within '{current_file}'."
    )

    VALID = [
        # Everything is allowed
        Valid("import common"),
        Valid(
            "import common",
            config=_gen_testcase_config({"some_dir": {"rules": [["*", "allow"]]}}),
            filename="some_dir/file.py",
        ),
        # This import is allowlisted
        Valid(
            "import common",
            config=_gen_testcase_config(
                {"some_dir": {"rules": [["common", "allow"], ["*", "deny"]]}}
            ),
            filename="some_dir/file.py",
        ),
        # Allow children of a allowlisted module
        Valid(
            "from common.foo import bar",
            config=_gen_testcase_config(
                {"some_dir": {"rules": [["common", "allow"], ["*", "deny"]]}}
            ),
            filename="some_dir/file.py",
        ),
        # Validate rules are evaluted in order
        Valid(
            "from common.foo import bar",
            config=_gen_testcase_config(
                {
                    "some_dir": {
                        "rules": [
                            ["common.foo.bar", "allow"],
                            ["common", "deny"],
                            ["*", "deny"],
                        ]
                    }
                }
            ),
            filename="some_dir/file.py",
        ),
        # Built-in modules are fine
        Valid(
            "import ast",
            config=_gen_testcase_config({"some_dir": {"rules": [["*", "deny"]]}}),
            filename="some_dir/file.py",
        ),
        # Relative imports
        Valid(
            "from . import module",
            config=_gen_testcase_config(
                {".": {"rules": [["common.safe", "allow"], ["*", "deny"]]}}
            ),
            filename="common/safe/file.py",
        ),
        Valid(
            "from ..safe import module",
            config=_gen_testcase_config(
                {"common": {"rules": [["common.safe", "allow"], ["*", "deny"]]}}
            ),
            filename="common/unsafe/file.py",
        ),
        # Ignore some relative module that leaves the repo root
        Valid(
            "from ....................................... import module",
            config=_gen_testcase_config({".": {"rules": [["*", "deny"]]}}),
            filename="file.py",
        ),
        # File belongs to more than one directory setting (should enforce closest parent directory)
        Valid(
            "from common.foo import bar",
            config=_gen_testcase_config(
                {
                    "dir_1/dir_2": {
                        "rules": [["common.foo.bar", "allow"], ["*", "deny"]]
                    },
                    "dir_1": {"rules": [["common.foo.bar", "deny"], ["*", "deny"]]},
                }
            ),
            filename="dir_1/dir_2/file.py",
        ),
        # File belongs to more than one directory setting, flipped order (should enforce closest parent directory)
        Valid(
            "from common.foo import bar",
            config=_gen_testcase_config(
                {
                    "dir_1": {"rules": [["common.foo.bar", "deny"], ["*", "deny"]]},
                    "dir_1/dir_2": {
                        "rules": [["common.foo.bar", "allow"], ["*", "deny"]]
                    },
                }
            ),
            filename="dir_1/dir_2/file.py",
        ),
        # Deny global
        Valid(
            """
            def local_scope():
                import common
            """,
            config=_gen_testcase_config({"dir": {"rules": [["*", "deny_global"]]}}),
            filename="dir/file.py",
        ),
        # Deny local
        Valid(
            """
            import common
            """,
            config=_gen_testcase_config({"dir": {"rules": [["*", "deny_local"]]}}),
            filename="dir/file.py",
        ),
        # Allow global
        Valid(
            """
            import common
            """,
            config=_gen_testcase_config({"dir": {"rules": [["*", "allow_global"]]}}),
            filename="dir/file.py",
        ),
        # Allow local
        Valid(
            """
            def local_scope():
                import common
            """,
            config=_gen_testcase_config({"dir": {"rules": [["*", "allow_local"]]}}),
            filename="dir/file.py",
        ),
    ]

    INVALID = [
        # Everything is denied
        Invalid(
            "import common",
            config=_gen_testcase_config({"some_dir": {"rules": [["*", "deny"]]}}),
            filename="some_dir/file.py",
        ),
        # Glob for a specific filename
        Invalid(
            "import common",
            config=_gen_testcase_config({"*/file.py": {"rules": [["*", "deny"]]}}),
            filename="some_dir/file.py",
        ),
        # Validate rules are evaluated in order
        Invalid(
            "from common.foo import bar",
            config=_gen_testcase_config(
                {
                    "some_dir": {
                        "rules": [
                            ["common.foo.bar", "deny"],
                            ["common", "allow"],
                            ["*", "allow"],
                        ]
                    }
                }
            ),
            filename="some_dir/file.py",
        ),
        # We should match against the real name, not the aliased name
        Invalid(
            "import common as not_common",
            config=_gen_testcase_config(
                {"some_dir": {"rules": [["common", "deny"], ["*", "allow"]]}}
            ),
            filename="some_dir/file.py",
        ),
        Invalid(
            "from common import bar as not_bar",
            config=_gen_testcase_config(
                {"some_dir": {"rules": [["common.bar", "deny"], ["*", "allow"]]}}
            ),
            filename="some_dir/file.py",
        ),
        # Relative imports
        Invalid(
            "from . import b",
            config=_gen_testcase_config({"common": {"rules": [["*", "deny"]]}}),
            filename="common/a.py",
        ),
        # File belongs to more than one directory setting, import from
        Invalid(
            "from common.foo import bar",
            config=_gen_testcase_config(
                {
                    "dir_1/dir_2": {"rules": [["*", "deny"]]},
                    "dir_1": {
                        "rules": [["common.foo.bar", "allow"], ["*", "deny"]],
                    },
                }
            ),
            filename="dir_1/dir_2/file.py",
        ),
        # File belongs to more than one directory setting, import
        Invalid(
            "import common",
            config=_gen_testcase_config(
                {
                    "dir_1/dir_2": {"rules": [["*", "deny"]]},
                    "dir_1": {
                        "rules": [["common", "allow"], ["*", "deny"]],
                    },
                }
            ),
            filename="dir_1/dir_2/file.py",
        ),
        # Deny global
        Invalid(
            """
            import common
            """,
            config=_gen_testcase_config({"dir": {"rules": [["*", "deny_global"]]}}),
            filename="dir/file.py",
        ),
        # Deny local
        Invalid(
            """
            def local_scope():
                import common
            """,
            config=_gen_testcase_config({"dir": {"rules": [["*", "deny_local"]]}}),
            filename="dir/file.py",
        ),
        # Allow global
        Invalid(
            """
            def local_scope():
                import common
            """,
            config=_gen_testcase_config({"dir": {"rules": [["*", "allow_global"]]}}),
            filename="dir/file.py",
        ),
        # Allow local
        Invalid(
            """
            import common
            """,
            config=_gen_testcase_config({"dir": {"rules": [["*", "allow_local"]]}}),
            filename="dir/file.py",
        ),
        # Custom message
        Invalid(
            """
            import common
            """,
            config=_gen_testcase_config(
                {
                    "dir": {
                        "rules": [["*", "deny"]],
                        "message": "'{imported}' cannot be imported from '{current_file}'",
                    }
                }
            ),
            filename="dir/file.py",
            expected_message=f"'common' cannot be imported from 'dir{os.path.sep}file.py'",
        ),
    ]

    def __init__(self, context: CstContext) -> None:
        super().__init__(context)
        self._repo_root = Path(self.context.config.repo_root).resolve()
        self._config = None
        self._abs_file_path = (self._repo_root / context.file_path).resolve()
        import_constraints_config = self.context.config.rule_config.get(
            self.__class__.__name__, None
        )
        if import_constraints_config is not None:
            formatted_config: Dict[
                Path, Dict[object, object]
            ] = self._parse_and_format_config(import_constraints_config)
            # Try matching the filepath and all of its parents paths in any of the
            # given configurations' glob patterns, stopping early if a match is
            # found in the config. The closest match in the settings will be used.
            parts = list(self._abs_file_path.parts)
            while parts:
                part_dir = os.path.join(*parts)
                parts.pop()
                for pattern, settings_for_dir in formatted_config.items():
                    if fnmatch(part_dir, str(pattern)):
                        self._config = _ImportConfig.from_config(settings_for_dir)
                        break
                if self._config is not None:
                    break
        self._type_checking_stack = []

    def _parse_and_format_config(
        self, import_constraints_config: Dict[str, object]
    ) -> Dict[Path, Dict[object, object]]:
        # Normalizes paths, and converts all paths to absolute paths using the specified repo_root.
        formatted_config: Dict[Path, Dict[object, object]] = {}
        for dirname, dir_settings in import_constraints_config.items():
            abs_dirpath: Optional[Path] = None

            # If it's an absolute path, make sure it's relative to repo_root (which should be an absolute path).
            if os.path.isabs(dirname):
                if dirname.startswith(str(self._repo_root)):
                    abs_dirpath = Path(dirname)
            # Otherwise assume all relative paths exist under repo_root, and don't add paths that leave repo_root (eg: '../path')
            else:
                abs_dirname = os.path.normpath(os.path.join(self._repo_root, dirname))
                if abs_dirname.startswith(str(self._repo_root)):
                    abs_dirpath = Path(abs_dirname)

            if not isinstance(dir_settings, dict):
                raise ValueError(
                    f"Invalid entry `{dir_settings}`.\n"
                    + "You must specify settings in key-value format under a directory."
                )
            if abs_dirpath is not None:
                formatted_config[abs_dirpath] = dir_settings

        return formatted_config

    def should_skip_file(self) -> bool:
        config = self._config
        return config is None or (config.ignore_tests and self.context.in_tests)

    def visit_If(self, node: cst.If) -> None:
        # TODO: Handle stuff like typing.TYPE_CHECKING
        test = node.test
        if isinstance(test, cst.Name) and test.value == "TYPE_CHECKING":
            self._type_checking_stack.append(node)

    def leave_If(self, original_node: cst.If) -> None:
        if self._type_checking_stack and self._type_checking_stack[-1] is original_node:
            self._type_checking_stack.pop()

    def visit_Import(self, node: cst.Import) -> None:
        self._check_names(
            node, (get_full_name_for_node_or_raise(alias.name) for alias in node.names)
        )

    def visit_ImportFrom(self, node: cst.ImportFrom) -> None:
        module = node.module
        abs_module = self._to_absolute_module(
            get_full_name_for_node_or_raise(module) if module is not None else "",
            len(node.relative),
        )
        if abs_module is not None:
            names = node.names
            if isinstance(names, Sequence):
                self._check_names(
                    node, (f"{abs_module}.{alias.name.value}" for alias in names)
                )

    def _to_absolute_module(self, module: Optional[str], level: int) -> Optional[str]:
        if level == 0:
            return module

        # Get the absolute path of the file
        current_dir = self._abs_file_path
        for __ in range(level):
            current_dir = current_dir.parent

        if (
            current_dir != self._repo_root
            and self._repo_root not in current_dir.parents
        ):
            return None

        prefix = ".".join(current_dir.relative_to(self._repo_root).parts)
        return f"{prefix}.{module}" if module is not None else prefix

    def _check_names(self, node: cst.CSTNode, names: Iterable[str]) -> None:
        config = self._config
        if config is None or (config.ignore_types and self._type_checking_stack):
            return
        message = config.message or self.MESSAGE
        for name in names:
            if name.split(".", 1)[0] not in _get_local_roots(self._repo_root):
                continue
            rule = config.match(name)
            action = rule.action
            if action in {
                RuleAction.ALLOW_GLOBAL,
                RuleAction.ALLOW_LOCAL,
                RuleAction.DENY_GLOBAL,
                RuleAction.DENY_LOCAL,
            }:
                scope = self.get_metadata(ScopeProvider, node)
                if isinstance(scope, GlobalScope):
                    if action in {RuleAction.ALLOW_GLOBAL, RuleAction.DENY_LOCAL}:
                        continue
                else:
                    if action in {RuleAction.ALLOW_LOCAL, RuleAction.DENY_GLOBAL}:
                        continue
            elif action == RuleAction.ALLOW:
                continue
            self.report(
                node,
                message.format(
                    imported=name,
                    current_file=self.context.file_path,
                ),
            )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from collections import defaultdict
from dataclasses import dataclass
from itertools import chain, islice
from typing import Dict, Iterable, Mapping, Optional, Set, TYPE_CHECKING

from libcst.metadata import FullRepoManager, TypeInferenceProvider


if TYPE_CHECKING:
    from logging import Logger

    from libcst.metadata.base_provider import ProviderT


BATCH_SIZE: int = 100


PLACEHOLDER_CACHES: Dict["ProviderT", object] = {TypeInferenceProvider: {"types": []}}


@dataclass(frozen=True)
class FullRepoMetadataConfig:
    providers: Set["ProviderT"]
    timeout_seconds: int
    repo_root_dir: str = ""
    batch_size: int = BATCH_SIZE
    logger: Optional["Logger"] = None


def get_repo_caches(
    paths: Iterable[str],
    config: FullRepoMetadataConfig,
) -> Mapping[str, Dict["ProviderT", object]]:
    """
    Generate type metadata by instantiating a :class:`~libcst.metadata.FullRepoManager` with
    :class:`~libcst.metadata.FullRepoManager` passed to ```providers``` parameter.

    :param paths: An iterable of paths to files to pass to :class:`~libcst.metadata.FullRepoManager` constructor's
    ```paths``` argument. These will be split in batches where the combined length of each path in the batch is <=
    ```arg_size```.

    :param timeout: The number of seconds at which to cap the pyre query which is run as a subprocess during cache resolving.

    :param repo_root_dir: Root directory of paths in ```paths```.

    :param batch_size: The size of the batch of paths to pass in each call to the :class:`~libcst.metadata.FullRepoManager` constructor.
    """
    caches = {}
    paths_iter = iter(paths)
    head: Optional[str] = next(paths_iter, None)
    while head is not None:
        paths_batch = tuple(chain([head], islice(paths_iter, config.batch_size - 1)))
        head = next(paths_iter, None)
        frm = FullRepoManager(
            repo_root_dir=config.repo_root_dir,
            paths=paths_batch,
            providers=config.providers,
            timeout=config.timeout_seconds,
        )
        try:
            frm.resolve_cache()
        except Exception:
            # We want to fail silently since some metadata providers can be flaky. If a logger is provided by the caller, we'll add a log here.
            logger = config.logger
            if logger is not None:
                logger.warning(
                    "Failed to retrieve metadata cache.",
                    exc_info=True,
                    extra={"paths": paths_batch},
                )
            # Populate with placeholder caches to avoid failures down the line. This will however result in reduced functionality in cache-dependent lint rules.
            caches.update(
                dict.fromkeys(
                    paths_batch,
                    {
                        provider: PLACEHOLDER_CACHES[provider]
                        for provider in config.providers
                    },
                )
            )
        else:
            # TODO: remove access of private variable when public `cache` property is available in libcst.metadata.FullRepoManager API.
            batch_caches = defaultdict(dict)
            for provider, files in frm._cache.items():
                for _path, cache in files.items():
                    batch_caches[_path][provider] = cache
            caches.update(batch_caches)
    return caches



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import libcst as cst
import libcst.matchers as m

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


UNNECESSARY_LITERAL: str = (
    "It's unnecessary to use a list or tuple within a call to {func} since"
    + " there is literal syntax for this type"
)
UNNCESSARY_CALL: str = (
    "It's slower to call {func}() than using the empty literal, because"
    + " the name {func} must be looked up in the global scope in case it has"
    + " been rebound."
)


class RewriteToLiteralRule(CstLintRule):
    """
    A derivative of flake8-comprehensions' C405-C406 and C409-C410. It's
    unnecessary to use a list or tuple literal within a call to tuple, list,
    set, or dict since there is literal syntax for these types.
    """

    VALID = [
        Valid("(1, 2)"),
        Valid("()"),
        Valid("[1, 2]"),
        Valid("[]"),
        Valid("{1, 2}"),
        Valid("set()"),
        Valid("{1: 2, 3: 4}"),
        Valid("{}"),
    ]

    INVALID = [
        Invalid("tuple([1, 2])", expected_replacement="(1, 2)"),
        Invalid("tuple((1, 2))", expected_replacement="(1, 2)"),
        Invalid("tuple([])", expected_replacement="()"),
        Invalid("list([1, 2, 3])", expected_replacement="[1, 2, 3]"),
        Invalid("list((1, 2, 3))", expected_replacement="[1, 2, 3]"),
        Invalid("list([])", expected_replacement="[]"),
        Invalid("set([1, 2, 3])", expected_replacement="{1, 2, 3}"),
        Invalid("set((1, 2, 3))", expected_replacement="{1, 2, 3}"),
        Invalid("set([])", expected_replacement="set()"),
        Invalid(
            "dict([(1, 2), (3, 4)])",
            expected_replacement="{1: 2, 3: 4}",
        ),
        Invalid(
            "dict(((1, 2), (3, 4)))",
            expected_replacement="{1: 2, 3: 4}",
        ),
        Invalid(
            "dict([[1, 2], [3, 4], [5, 6]])",
            expected_replacement="{1: 2, 3: 4, 5: 6}",
        ),
        Invalid("dict([])", expected_replacement="{}"),
        Invalid("tuple()", expected_replacement="()"),
        Invalid("list()", expected_replacement="[]"),
        Invalid("dict()", expected_replacement="{}"),
    ]

    def visit_Call(self, node: cst.Call) -> None:
        if m.matches(
            node,
            m.Call(
                func=m.Name("tuple") | m.Name("list") | m.Name("set") | m.Name("dict"),
                args=[m.Arg(value=m.List() | m.Tuple())],
            ),
        ) or m.matches(
            node,
            m.Call(func=m.Name("tuple") | m.Name("list") | m.Name("dict"), args=[]),
        ):

            pairs_matcher = m.ZeroOrMore(
                m.Element(m.Tuple(elements=[m.DoNotCare(), m.DoNotCare()]))
                | m.Element(m.List(elements=[m.DoNotCare(), m.DoNotCare()]))
            )

            exp = cst.ensure_type(node, cst.Call)
            call_name = cst.ensure_type(exp.func, cst.Name).value

            # If this is a empty call, it's an Unnecessary Call where we rewrite the call
            # to literal, except set().
            if not exp.args:
                elements = []
                message_formatter = UNNCESSARY_CALL
            else:
                arg = exp.args[0].value
                elements = cst.ensure_type(
                    arg, cst.List if isinstance(arg, cst.List) else cst.Tuple
                ).elements
                message_formatter = UNNECESSARY_LITERAL

            if call_name == "tuple":
                new_node = cst.Tuple(elements=elements)
            elif call_name == "list":
                new_node = cst.List(elements=elements)
            elif call_name == "set":
                # set() doesn't have an equivelant literal call. If it was
                # matched here, it's an unnecessary literal suggestion.
                if len(elements) == 0:
                    self.report(
                        node,
                        UNNECESSARY_LITERAL.format(func=call_name),
                        replacement=node.deep_replace(
                            node, cst.Call(func=cst.Name("set"))
                        ),
                    )
                    return
                new_node = cst.Set(elements=elements)
            elif len(elements) == 0 or m.matches(
                exp.args[0].value,
                m.Tuple(elements=[pairs_matcher]) | m.List(elements=[pairs_matcher]),
            ):
                new_node = cst.Dict(
                    elements=[
                        (
                            lambda val: cst.DictElement(
                                val.elements[0].value, val.elements[1].value
                            )
                        )(
                            cst.ensure_type(
                                ele.value,
                                cst.Tuple
                                if isinstance(ele.value, cst.Tuple)
                                else cst.List,
                            )
                        )
                        for ele in elements
                    ]
                )
            else:
                # Unrecoginized form
                return

            self.report(
                node,
                message_formatter.format(func=call_name),
                replacement=node.deep_replace(node, new_node),
            )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from libcst.testing.utils import UnitTest

from fixit.common.base import _get_code


class GetCodeTest(UnitTest):
    def test_validate_message_old_code(self) -> None:
        # lint-ignore: UseClassNameAsCodeRule
        self.assertEqual(_get_code("IG00 Old-school message", "SomeClassName"), "IG00")

    def test_return_classname(self) -> None:
        self.assertEqual(
            _get_code("Message blah blah", "SomeClassName"), "SomeClassName"
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path

from libcst.testing.utils import UnitTest

from fixit.common.base import LintConfig
from fixit.common.pseudo_rule import PseudoContext
from fixit.rule_lint_engine import lint_file
from fixit.rules.flake8_compat import Flake8PseudoLintRule


class Flake8PseudoLintRuleTest(UnitTest):
    def test_lint_file(self) -> None:
        context = PseudoContext(
            file_path=Path("dummy/file/path.py"), source=b"undefined_fn()\n"
        )
        rule = Flake8PseudoLintRule(context)
        results = list(rule.lint_file())
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].code, "F821")  # undefined name

    def test_flake8_noqa_is_disabled(self) -> None:
        context = PseudoContext(
            file_path=Path("dummy/file/path.py"), source=b"undefined_fn()  # noqa\n"
        )
        rule = Flake8PseudoLintRule(context)
        results = list(rule.lint_file())
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].code, "F821")  # undefined name

    def test_lint_file_with_framework(self) -> None:
        results = list(
            lint_file(
                file_path=Path("dummy/file/path.py"),
                source=b"undefined_fn()\n",
                rules={Flake8PseudoLintRule},
                config=LintConfig(),
            )
        )
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].code, "F821")  # undefined name

    def test_lint_ignore_with_framework(self) -> None:
        results = list(
            lint_file(
                file_path=Path("dummy/file/path.py"),
                source=b"# lint-ignore: F821: testing ignores\nundefined_fn()\n",
                rules={Flake8PseudoLintRule},
                config=LintConfig(),
            )
        )
        self.assertEqual(results, [])



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import ast
import io
import tokenize
from pathlib import Path
from typing import Iterable

from libcst.testing.utils import UnitTest

from fixit.common.base import LintConfig
from fixit.common.pseudo_rule import PseudoContext, PseudoLintRule
from fixit.common.report import BaseLintRuleReport
from fixit.rule_lint_engine import lint_file


DUMMY_FILE_PATH = Path(__file__)
DUMMY_SOURCE = b"pass\npass\npass\n"
DUMMY_LINT_CODE = "DummyLintRule"
DUMMY_LINT_MESSAGE = "dummy lint message"


class PseudoContextTest(UnitTest):
    def setUp(self) -> None:
        self.dummy_tokens = tuple(tokenize.tokenize(io.BytesIO(DUMMY_SOURCE).readline))
        self.dummy_ast_tree = ast.parse(DUMMY_SOURCE)

    def test_tokens(self) -> None:
        full_context = PseudoContext(
            file_path=DUMMY_FILE_PATH, source=DUMMY_SOURCE, tokens=self.dummy_tokens
        )
        self.assertIs(full_context.tokens, self.dummy_tokens)
        partial_context = PseudoContext(file_path=DUMMY_FILE_PATH, source=DUMMY_SOURCE)
        self.assertEqual(partial_context.tokens, self.dummy_tokens)
        self.assertIsNot(partial_context.tokens, self.dummy_tokens)

    def test_ast_tree(self) -> None:
        full_context = PseudoContext(
            file_path=DUMMY_FILE_PATH, source=DUMMY_SOURCE, ast_tree=self.dummy_ast_tree
        )
        self.assertIs(full_context.ast_tree, self.dummy_ast_tree)
        partial_context = PseudoContext(file_path=DUMMY_FILE_PATH, source=DUMMY_SOURCE)
        # partial_context.ast_tree should be equivalent to self.dummy_ast_tree
        self.assertIsNot(partial_context.ast_tree, self.dummy_ast_tree)


class PseudoLintRuleTest(UnitTest):
    def test_pseudo_lint_rule(self) -> None:
        class DummyLintRuleReport(BaseLintRuleReport):
            pass

        dummy_report = DummyLintRuleReport(
            file_path=DUMMY_FILE_PATH,
            code=DUMMY_LINT_CODE,
            message=DUMMY_LINT_MESSAGE,
            line=1,
            column=0,
        )

        class DummyPseudoLintRule(PseudoLintRule):
            def lint_file(self) -> Iterable[BaseLintRuleReport]:
                return [dummy_report]

        reports = lint_file(
            DUMMY_FILE_PATH,
            DUMMY_SOURCE,
            config=LintConfig(),
            rules={DummyPseudoLintRule},
        )
        self.assertEqual(reports, [dummy_report])



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from fixit.common.base import CstLintRule


# Dummy rule for integration testing purposes.
class DummyRule1(CstLintRule):
    pass



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Iterable

from libcst.testing.utils import data_provider, UnitTest

from fixit.common.insert_suppressions import (
    insert_suppressions,
    SuppressionComment,
    SuppressionCommentKind,
)
from fixit.common.utils import dedent_with_lstrip


class InsertSuppressionsTest(UnitTest):
    @data_provider(
        {
            "simple_fixme": {
                "before": dedent_with_lstrip(
                    """
                    def fn():
                        ...
                    """
                ),
                "after": dedent_with_lstrip(
                    """
                    # lint-fixme: IgnoredRule: Some message
                    def fn():
                        ...
                    """
                ),
                "comments": [
                    SuppressionComment(
                        kind=SuppressionCommentKind.FIXME,
                        before_line=1,
                        code="IgnoredRule",
                        message="Some message",
                    )
                ],
            },
            "simple_ignore": {
                "before": dedent_with_lstrip(
                    """
                    def fn():
                        ...
                    """
                ),
                "after": dedent_with_lstrip(
                    """
                    # lint-ignore: IgnoredRule: Some message
                    def fn():
                        ...
                    """
                ),
                "comments": [
                    SuppressionComment(
                        kind=SuppressionCommentKind.IGNORE,
                        before_line=1,
                        code="IgnoredRule",
                        message="Some message",
                    )
                ],
            },
            "no_message": {
                "before": dedent_with_lstrip(
                    """
                    def fn():
                        ...
                    """
                ),
                "after": dedent_with_lstrip(
                    """
                    # lint-fixme: IgnoredRule
                    def fn():
                        ...
                    """
                ),
                "comments": [
                    SuppressionComment(
                        kind=SuppressionCommentKind.FIXME,
                        before_line=1,
                        code="IgnoredRule",
                    )
                ],
            },
            "indented": {
                "before": dedent_with_lstrip(
                    """
                    def fn():
                        ...
                    """
                ),
                "after": dedent_with_lstrip(
                    """
                    def fn():
                        # lint-fixme: IgnoredRule: Some message
                        ...
                    """
                ),
                "comments": [
                    SuppressionComment(
                        kind=SuppressionCommentKind.FIXME,
                        before_line=2,
                        code="IgnoredRule",
                        message="Some message",
                    )
                ],
            },
            "indented_tabs": {
                "before": dedent_with_lstrip(
                    """
                    def fn():
                    \t...
                    """
                ),
                "after": dedent_with_lstrip(
                    """
                    def fn():
                    \t# lint-fixme: IgnoredRule: Some message
                    \t...
                    """
                ),
                "comments": [
                    SuppressionComment(
                        kind=SuppressionCommentKind.FIXME,
                        before_line=2,
                        code="IgnoredRule",
                        message="Some message",
                    )
                ],
            },
            "multiple_comments": {
                "before": dedent_with_lstrip(
                    """
                    def fn():
                        ...
                    """
                ),
                "after": dedent_with_lstrip(
                    """
                    # lint-fixme: IgnoredRule: Some message
                    # lint-fixme: IgnoredRule1: Another message
                    def fn():
                        # lint-fixme: IgnoredRule2: Yet another
                        ...
                    """
                ),
                "comments": [
                    SuppressionComment(
                        kind=SuppressionCommentKind.FIXME,
                        before_line=1,
                        code="IgnoredRule",
                        message="Some message",
                    ),
                    SuppressionComment(
                        kind=SuppressionCommentKind.FIXME,
                        before_line=1,
                        code="IgnoredRule1",
                        message="Another message",
                    ),
                    SuppressionComment(
                        kind=SuppressionCommentKind.FIXME,
                        before_line=2,
                        code="IgnoredRule2",
                        message="Yet another",
                    ),
                ],
            },
            "multiline_comment": {
                "before": dedent_with_lstrip(
                    """
                    def fn():
                        ...
                    """
                ),
                "after": dedent_with_lstrip(
                    """
                    def fn():
                        # lint-fixme: IgnoredRule:
                        # lint: Some really long
                        # lint: message that
                        # lint: rambles on and on
                        # lint: that needs to be
                        # lint: wrapped
                        ...
                    """
                ),
                "comments": [
                    SuppressionComment(
                        kind=SuppressionCommentKind.FIXME,
                        before_line=2,
                        code="IgnoredRule",
                        message=(
                            "Some really long message that rambles on and on that "
                            + "needs to be wrapped"
                        ),
                        max_lines=(2**32),
                    )
                ],
                "code_width": 30,
            },
            "newlines_in_message": {
                "before": dedent_with_lstrip(
                    """
                    def fn():
                        ...
                    """
                ),
                "after": dedent_with_lstrip(
                    """
                    def fn():
                        # lint-fixme: IgnoredRule: This is the first line.
                        # lint: This is a subsequent line followed by a blank line.
                        # lint:
                        # lint: And this is the last line.
                        ...
                    """
                ),
                "comments": [
                    SuppressionComment(
                        kind=SuppressionCommentKind.FIXME,
                        before_line=2,
                        code="IgnoredRule",
                        message=(
                            "This is the first line.\n"
                            + "This is a subsequent line followed by a blank line.\n"
                            + "\n"
                            + "And this is the last line."
                        ),
                        max_lines=(2**32),
                    )
                ],
            },
            "logical_line_continuation": {
                "before": dedent_with_lstrip(
                    """
                    value = "abc"
                    value = \\
                        "abcd" + \\
                        "efgh" + \\
                        "ijkl" + \\
                        "mnop"
                    """
                ),
                "after": dedent_with_lstrip(
                    """
                    value = "abc"
                    # lint-fixme: IgnoredRule: Some message
                    value = \\
                        "abcd" + \\
                        "efgh" + \\
                        "ijkl" + \\
                        "mnop"
                    """
                ),
                "comments": [
                    # Line 4 isn't a logical line, so we expect that the comment will
                    # be put on the first logical line above it.
                    SuppressionComment(
                        kind=SuppressionCommentKind.FIXME,
                        before_line=4,
                        code="IgnoredRule",
                        message="Some message",
                    )
                ],
            },
            "logical_line_multiline_string": {
                "before": dedent_with_lstrip(
                    """
                    value = "abc"
                    value = '''
                        abcd
                        efgh
                        ijkl
                        mnop
                    '''
                    """
                ),
                "after": dedent_with_lstrip(
                    """
                    value = "abc"
                    # lint-fixme: IgnoredRule: Some message
                    value = '''
                        abcd
                        efgh
                        ijkl
                        mnop
                    '''
                    """
                ),
                "comments": [
                    # Line 4 isn't a logical line, so we expect that the comment will
                    # be put on the first logical line above it.
                    SuppressionComment(
                        kind=SuppressionCommentKind.FIXME,
                        before_line=4,
                        code="IgnoredRule",
                        message="Some message",
                    )
                ],
            },
            "max_lines_first_block": {
                "before": dedent_with_lstrip(
                    """
                    def fn():
                        ...
                    """
                ),
                "after": dedent_with_lstrip(
                    """
                    # lint-fixme: IgnoredRule: first block ...
                    def fn():
                        ...
                    """
                ),
                "comments": [
                    SuppressionComment(
                        kind=SuppressionCommentKind.FIXME,
                        before_line=1,
                        code="IgnoredRule",
                        message="first block\n\nsecond block\nthird block",
                        max_lines=1,
                    )
                ],
            },
            "max_lines_between_blocks": {
                "before": dedent_with_lstrip(
                    """
                    def fn():
                        ...
                    """
                ),
                "after": dedent_with_lstrip(
                    """
                    # lint-fixme: IgnoredRule: first block
                    # lint: ...
                    def fn():
                        ...
                    """
                ),
                "comments": [
                    SuppressionComment(
                        kind=SuppressionCommentKind.FIXME,
                        before_line=1,
                        code="IgnoredRule",
                        message="first block\n\nsecond block\nthird block",
                        max_lines=2,
                    )
                ],
            },
            "max_lines_subsequent_blocks": {
                "before": dedent_with_lstrip(
                    """
                    def fn():
                        ...
                    """
                ),
                "after": dedent_with_lstrip(
                    """
                    # lint-fixme: IgnoredRule: first block
                    # lint:
                    # lint: second block ...
                    def fn():
                        ...
                    """
                ),
                "comments": [
                    SuppressionComment(
                        kind=SuppressionCommentKind.FIXME,
                        before_line=1,
                        code="IgnoredRule",
                        message="first block\n\nsecond block\nthird block",
                        max_lines=3,
                    )
                ],
            },
            # In this example the last visible line wouldn't normally need to be
            # truncated, but we don't quite have enough space for the "[...]" ellipsis
            # at the end.
            "max_lines_requires_trimming": {
                "before": dedent_with_lstrip(
                    """
                    def fn():
                        ...
                    """
                ),
                "after": dedent_with_lstrip(
                    """
                    # lint-fixme: IgnoredRule: first line
                    # lint: second line which is too ...
                    def fn():
                        ...
                    """
                ),
                "comments": [
                    SuppressionComment(
                        kind=SuppressionCommentKind.FIXME,
                        before_line=1,
                        code="IgnoredRule",
                        message="first line\nsecond line which is too long\nlast line",
                        max_lines=2,
                    )
                ],
                "code_width": 40,  # the truncated comment is 38 characters long (<40)
            },
        }
    )
    def test_insert_suppressions(
        self,
        *,
        before: str,
        after: str,
        comments: Iterable[SuppressionComment],
        code_width: int = 1000,
        min_comment_width: int = 1,
    ) -> None:
        result = insert_suppressions(
            before.encode("utf-8"),
            comments,
            code_width=code_width,
            min_comment_width=min_comment_width,
        )
        updated_source = result.updated_source.decode("utf-8")
        self.assertEqual(updated_source, after)
        self.assertEqual(len(result.failed_insertions), 0)



# Copyright (c) Meta Platforms, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from libcst._add_slots import add_slots
from libcst.metadata import CodeRange

FileContent = bytes


@dataclass
class Options:
    """
    Command-line options to affect runtime behavior
    """

    debug: Optional[bool]
    config_file: Optional[Path]


@dataclass
class Config:
    """
    Materialized configuration valid for processing a single file
    """

    path: Path


@add_slots
@dataclass(frozen=True)
class LintViolation:
    rule_name: str
    range: CodeRange
    message: str
    autofixable: bool


@dataclass
class Result:
    """
    A single lint result for a given file and lint rule
    """

    path: Path
    violation: Optional[LintViolation]
    error: Optional[Exception] = None



import functools
import logging
import time
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import replace
from typing import (
    Callable,
    ClassVar,
    Collection,
    ContextManager,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Set,
    TypeVar,
    Union,
)

from libcst import (
    BatchableCSTVisitor,
    CSTNode,
    FlattenSentinel,
    parse_module,
    RemovalSentinel,
)
from libcst.metadata import (
    CodePosition,
    CodeRange,
    MetadataWrapper,
    PositionProvider,
    ProviderT,
)

from fixit.types import FileContent, LintViolation
from . import LintRule, LintRunner, TimingsHook


VisitorMethod = Callable[[CSTNode], None]
VisitHook = Callable[[str], ContextManager]

logger = logging.getLogger(__name__)


class CSTLintRunner(LintRunner["CSTLintRule"]):
    def collect_violations(
        self,
        source: FileContent,
        rules: Collection["CSTLintRule"],
        timings_hook: Optional[TimingsHook] = None,
    ) -> Iterable[LintViolation]:
        @contextmanager
        def visit_hook(name: str) -> Iterator[None]:
            start = time.perf_counter()
            try:
                yield
            finally:
                duration_us = int(1000 * 1000 * (time.perf_counter() - start))
                logger.debug(f"PERF: {name} took {duration_us} µs")
                self.timings[name] += duration_us

        for rule in rules:
            rule._visit_hook = visit_hook

        mod = MetadataWrapper(parse_module(source), unsafe_skip_copy=True)
        mod.visit_batched(rules)
        for rule in rules:
            for violation in rule._violations:
                yield violation
        if timings_hook:
            timings_hook(self.timings)


class CSTLintRule(LintRule, BatchableCSTVisitor):
    METADATA_DEPENDENCIES: ClassVar[Collection[ProviderT]] = (PositionProvider,)

    _runner = CSTLintRunner
    _visit_hook: Optional[VisitHook] = None

    def report(
        self,
        node: CSTNode,
        message: Optional[str] = None,
        *,
        position: Optional[Union[CodePosition, CodeRange]] = None,
        replacement: Optional[Union[CSTNode, RemovalSentinel, FlattenSentinel]] = None,
    ) -> None:
        rule_name = type(self).__name__
        if not message:
            # backwards compat with Fixit 1.0 api
            message = getattr(self, "MESSAGE", None)
            if not message:
                raise ValueError(f"No message provided in {rule_name}")

        if position is None:
            position = self.get_metadata(PositionProvider, node)
        elif isinstance(position, CodePosition):
            end = replace(position, line=position.line + 1, column=0)
            position = CodeRange(start=position, end=end)
        self._violations.append(
            LintViolation(
                rule_name,
                range=position,
                message=message,
                autofixable=bool(replacement),
            )
        )

    def get_visitors(self) -> Mapping[str, VisitorMethod]:
        def _wrap(name: str, func: VisitorMethod) -> VisitorMethod:
            @functools.wraps(func)
            def wrapper(node: CSTNode) -> None:
                if self._visit_hook:
                    with self._visit_hook(name):
                        return func(node)
                return func(node)

            return wrapper

        return {
            name: _wrap(f"{type(self).__name__}.{name}", visitor)
            for (name, visitor) in super().get_visitors().items()
        }


CstLintRule = CSTLintRule



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from collections import defaultdict
from logging import getLogger, Handler, Logger, LogRecord
from subprocess import TimeoutExpired
from typing import DefaultDict, Iterable, List, Mapping, Type, TYPE_CHECKING

from libcst.metadata import TypeInferenceProvider

from fixit.cli import FullRepoMetadataConfig
from fixit.cli.utils import print_yellow
from fixit.common.base import CstLintRule
from fixit.common.full_repo_metadata import get_repo_caches
from fixit.common.utils import LintRuleCollectionT


if TYPE_CHECKING:
    from libcst.metadata.base_provider import ProviderT


class MetadataCacheErrorHandler(Handler):
    timeout_paths: List[str] = []
    other_exceptions: DefaultDict[Type[Exception], List[str]] = defaultdict(list)

    def emit(self, record: LogRecord) -> None:
        # According to logging documentation, exc_info will be a tuple of three values: (type, value, traceback)
        # see https://docs.python.org/3.8/library/logging.html#logrecord-objects
        exc_info = record.exc_info
        if exc_info is not None:
            exc_type = exc_info[0]
            failed_paths = record.__dict__.get("paths")
            if exc_type is not None:
                # Store exceptions in memory for processing later.
                if exc_type is TimeoutExpired:
                    self.timeout_paths += failed_paths
                else:
                    # pyre-fixme[6]: Expected `Type[Exception]` for 1st param but
                    #  got `Type[BaseException]`.
                    self.other_exceptions[exc_type] += failed_paths


def rules_require_metadata_cache(rules: LintRuleCollectionT) -> bool:
    return any(
        # pyre-fixme[16]: `PseudoLintRule` has no attribute `requires_metadata_caches`.
        issubclass(r, CstLintRule) and getattr(r, "requires_metadata_caches")()
        for r in rules
    )


def get_metadata_caches(
    cache_timeout: int, file_paths: Iterable[str]
) -> Mapping[str, Mapping["ProviderT", object]]:
    """
    Returns a metadata cache for each file in ``file_paths``.
    """
    logger: Logger = getLogger("Metadata Caches Logger")
    handler = MetadataCacheErrorHandler()
    logger.addHandler(handler)
    full_repo_metadata_config: FullRepoMetadataConfig = FullRepoMetadataConfig(
        providers={TypeInferenceProvider},
        timeout_seconds=cache_timeout,
        batch_size=100,
        logger=logger,
    )
    metadata_caches = get_repo_caches(file_paths, full_repo_metadata_config)
    # Let user know of any cache retrieval failures.
    if handler.timeout_paths:
        print(
            "Unable to get metadata cache for the following paths:\n"
            + "\n".join(handler.timeout_paths)
        )
        print_yellow("Try increasing the --cache-timeout value or passing fewer files.")
    for k, v in handler.other_exceptions.items():
        print(f"Encountered exception {k} for the following paths:\n" + "\n".join(v))
        print_yellow("Running `pyre start` may solve the issue.")
    return metadata_caches



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import re
from abc import ABCMeta
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Type, TYPE_CHECKING, Union

import libcst as cst
from libcst import BatchableCSTVisitor
from libcst.metadata import (
    BaseMetadataProvider,
    CodePosition,
    MetadataWrapper,
    PositionProvider,
    TypeInferenceProvider,
)

from fixit.common.report import BaseLintRuleReport, CstLintRuleReport


if TYPE_CHECKING:
    from libcst.metadata.base_provider import ProviderT

    from fixit.common.pseudo_rule import PseudoLintRule


LintRuleT = Union[Type["CstLintRule"], Type["PseudoLintRule"]]

CACHE_DEPENDENT_PROVIDERS: Tuple["ProviderT"] = (TypeInferenceProvider,)


def _get_code(message: str, class_name: str) -> str:
    """Extract the lint code from the beginning of the lint message."""
    # TODO: Deprecate this function.
    code_match = re.match(r"^(?P<code>IG\d+) \S", message)
    if not code_match:
        return class_name
    return code_match.group("code")


DEFAULT_PACKAGES = ["fixit.rules"]
DEFAULT_PATTERNS = [f"@ge{''}nerated", "@nolint"]


@dataclass(frozen=True)
class LintConfig:
    allow_list_rules: List[str] = field(default_factory=list)
    block_list_patterns: List[str] = field(default_factory=lambda: DEFAULT_PATTERNS)
    block_list_rules: List[str] = field(default_factory=list)
    fixture_dir: str = "./fixtures"
    use_noqa: bool = False
    formatter: List[str] = field(default_factory=list)
    packages: List[str] = field(default_factory=lambda: DEFAULT_PACKAGES)
    repo_root: str = "."
    rule_config: Dict[str, Dict[str, object]] = field(default_factory=dict)


class BaseContext:
    file_path: Path
    config: LintConfig
    reports: List[BaseLintRuleReport]

    def __init__(self, file_path: Path, config: LintConfig) -> None:
        self.file_path = file_path
        self.config = config
        self.reports = []

    @property
    def in_tests(self) -> bool:
        return self.file_path.name == "tests.py" or "tests" in self.file_path.parts

    @property
    def in_scripts(self) -> bool:
        return Path("distillery/scripts") in self.file_path.parents


class CstContext(BaseContext):
    wrapper: MetadataWrapper
    _source: bytes
    node_stack: List[cst.CSTNode]

    def __init__(
        self,
        wrapper: MetadataWrapper,
        source: bytes,
        file_path: Path,
        config: LintConfig,
    ) -> None:
        super().__init__(file_path, config)
        self.wrapper = wrapper
        # Keep the source around so we can use it in autofix diff generation. This is
        # private because lint rules should use the CST tree, not the source code. If we
        # exposed the source, it'd be providing rope for people to hang themselves with.
        self._source = source
        self.node_stack = []


class CstLintRule(BatchableCSTVisitor, metaclass=ABCMeta):
    #: a short message in one or two sentences show to user when the rule is violated.
    MESSAGE: Optional[str] = None

    METADATA_DEPENDENCIES: Tuple[Type[BaseMetadataProvider], ...] = (PositionProvider,)

    def __init__(self, context: CstContext) -> None:
        super().__init__()
        self.context = context

    def should_skip_file(self) -> bool:
        return False

    def report(
        self,
        node: cst.CSTNode,
        message: Optional[str] = None,
        *,
        position: Optional[CodePosition] = None,
        replacement: Optional[
            Union[cst.CSTNode, cst.RemovalSentinel, cst.FlattenSentinel]
        ] = None,
    ) -> None:
        """
        Report a lint violation for a given node. Optionally specify a custom
        position to report an error at or a replacement node for an auto-fix.
        """
        if position is None:
            position = self.context.wrapper.resolve(PositionProvider)[node].start

        if message is None:
            message = self.MESSAGE
            if message is None:
                raise Exception(f"No lint message was provided to rule: {self}")
        report = CstLintRuleReport(
            file_path=self.context.file_path,
            node=node,
            # TODO deprecate _get_code() completely and replace with self.__class__.__name__
            code=_get_code(message, self.__class__.__name__),
            message=message,
            line=position.line,
            # libcst columns are 0-indexed but arc is 1-indexed
            column=(position.column + 1),
            module=self.context.wrapper,
            module_bytes=self.context._source,
            replacement_node=replacement,
        )
        self.context.reports.append(report)

    @classmethod
    def requires_metadata_caches(cls) -> bool:
        return any(
            p in CACHE_DEPENDENT_PROVIDERS for p in cls.get_inherited_dependencies()
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
For backwards compatibility.
"""

from fixit.cli.formatter import LintRuleReportFormatter


__all__ = ["LintRuleReportFormatter"]



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import libcst as cst
import libcst.matchers as m

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


class NoAssertEqualsRule(CstLintRule):
    """
    Discourages use of ``assertEquals`` as it is deprecated (see https://docs.python.org/2/library/unittest.html#deprecated-aliases
    and https://bugs.python.org/issue9424). Use the standardized ``assertEqual`` instead.
    """

    MESSAGE: str = (
        '"assertEquals" is deprecated, use "assertEqual" instead.\n'
        + "See https://docs.python.org/2/library/unittest.html#deprecated-aliases and https://bugs.python.org/issue9424."
    )
    VALID = [Valid("self.assertEqual(a, b)")]
    INVALID = [
        Invalid(
            "self.assertEquals(a, b)",
            expected_replacement="self.assertEqual(a, b)",
        )
    ]

    def visit_Call(self, node: cst.Call) -> None:
        if m.matches(
            node,
            m.Call(func=m.Attribute(value=m.Name("self"), attr=m.Name("assertEquals"))),
        ):
            new_call = node.with_deep_changes(
                old_node=cst.ensure_type(node.func, cst.Attribute).attr,
                value="assertEqual",
            )
            self.report(node, replacement=new_call)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import re
import textwrap
import tokenize
from collections import deque
from dataclasses import dataclass
from enum import Enum
from io import BytesIO
from typing import Iterable, Mapping, Optional, Sequence

from fixit.common.line_mapping import LineMappingInfo


BODY_PREFIX = "# lint:"
BODY_PREFIX_WITH_SPACE = f"{BODY_PREFIX} "
MAX_LINES_PLACEHOLDER = " ..."
DEFAULT_CODE_WIDTH = 88
DEFAULT_MIN_COMMENT_WIDTH = 40


class SuppressionCommentKind(Enum):
    IGNORE = "lint-ignore"
    FIXME = "lint-fixme"


@dataclass(frozen=True, order=True)
class SuppressionComment:
    """
    Represents a `# lint-fixme` or `# lint-ignore` comment that prevents the linter from
    firing on the subsequent line. If the `message` is too long, the comment is wrapped
    to fill multiple lines.

    Wrapped lines are prefixed with `# lint:`, which allows us to figure out where a
    suppression comment starts and ends, so we can automate removal of unused multiline
    comments.

    At some point in the future, we may extend this to support `# lint-ignore` comments
    too.
    """

    kind: SuppressionCommentKind
    before_line: int  # 1-indexed
    code: str
    message: Optional[str] = None
    max_lines: int = 3

    def to_lines(self, width: int) -> Sequence[str]:
        message = self.message
        if message is None:
            return [f"# {self.kind.value}: {self.code}"]

        # chunk the message up splitting by newlines
        raw_message_lines = message.split("\n")
        initial_indent = f"# {self.kind.value}: {self.code}"

        lines = []
        lines.extend(
            textwrap.wrap(
                ": " + raw_message_lines[0],
                initial_indent=initial_indent,
                # We don't want a weirdly formed comment with the first line being
                # longer than the rest.
                width=max(len(initial_indent), width),
                subsequent_indent=BODY_PREFIX_WITH_SPACE,
                drop_whitespace=True,
            )
        )
        # textwrap replaces newlines (`\n`) with a space. This isn't the behavior we
        # want, so we need to wrap each line independently.
        for rml in raw_message_lines[1:]:
            if rml == "":
                lines.append(BODY_PREFIX)
            else:
                lines.extend(
                    textwrap.wrap(
                        rml,
                        width=width,
                        initial_indent=BODY_PREFIX_WITH_SPACE,
                        subsequent_indent=BODY_PREFIX_WITH_SPACE,
                    )
                )

        # Unfortunately our custom handling of newlines also means that we can't use
        # textwrap's `max_lines` and `placeholder` features, and we have to do it
        # ourselves.
        if len(lines) > self.max_lines:
            lines = lines[: self.max_lines]
            last_line = lines[-1]
            # keep removing words from the end until we have room for our placeholder
            while last_line and len(last_line) > (width - len(MAX_LINES_PLACEHOLDER)):
                # this must remove at least one character each time, otherwise we could
                # get stuck in an infinite loop
                last_line = re.sub(r"(\s*\S+|\s+)\Z", "", last_line)
            # if we removed too much, add the `# lint:` prefix back
            if len(last_line) < len(BODY_PREFIX):
                last_line = BODY_PREFIX
            last_line += MAX_LINES_PLACEHOLDER
            lines[-1] = last_line

        return lines


@dataclass(frozen=True)
class InsertSuppressionsResult:
    # An updated sequence of lines with included newlines. These lines can be joined to
    # generate the resulting source code.
    updated_source: bytes
    # It may not be possible to insert a comment where one was needed (e.g. we can't
    # insert a comment inside a multiline string). In those cases, we need to mark these
    # attempted insertions as failures.
    failed_insertions: Sequence[SuppressionComment]


def _get_indentations(tokens: Iterable[tokenize.TokenInfo]) -> Mapping[int, str]:
    """
    Maps logical lines to their indentations.

    This only works for logical lines because logical lines are the only lines with
    measurable indentation.
    """

    result = {}
    # We're 1-indexed, so 0 isn't a real line. The ENCODING dummy token uses it, but we
    # want to skip ENCODING.
    prev_line = 0
    for tok in tokens:
        if tok.type in (tokenize.INDENT, tokenize.DEDENT):
            # These dummy tokens have the wrong start column and always exist alongside
            # a non-dummy token. Skip them.
            continue
        # we only want to know about the first token on each unique line
        if tok.start[0] != prev_line:
            result[tok.start[0]] = tok.line[: tok.start[1]]
            prev_line = tok.start[0]
    return result


def insert_suppressions(
    source: bytes,
    comments: Iterable[SuppressionComment],
    *,
    code_width: int = DEFAULT_CODE_WIDTH,
    min_comment_width: int = DEFAULT_MIN_COMMENT_WIDTH,
) -> InsertSuppressionsResult:
    """
    Given an iterable of `lines`, forms a new sequence of lines with `comments`
    inserted.
    """
    encoding = tokenize.detect_encoding(BytesIO(source).readline)[0]
    tokens = tuple(tokenize.tokenize(BytesIO(source).readline))
    indentations = _get_indentations(tokens)
    physical_to_logical = LineMappingInfo.compute(tokens=tokens).physical_to_logical
    comments_queue = deque(sorted(comments))  # sort by line number
    updated_lines = []

    for line_number, line_bytes in enumerate(BytesIO(source).readlines(), start=1):
        while comments_queue:
            target_line = physical_to_logical[comments_queue[0].before_line]
            if target_line == line_number:
                indent = indentations[line_number]
                width = max(code_width - len(indent), min_comment_width)
                for line in comments_queue.popleft().to_lines(width):
                    updated_lines.append(f"{indent}{line}\n".encode(encoding))
            else:
                break
        updated_lines.append(line_bytes)

    return InsertSuppressionsResult(
        updated_source=b"".join(updated_lines), failed_insertions=tuple(comments_queue)
    )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import libcst as cst

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


class GatherSequentialAwaitRule(CstLintRule):
    """
    Discourages awaiting coroutines in a loop as this will run them sequentially. Using ``asyncio.gather()`` will run them concurrently.
    """

    MESSAGE: str = (
        "Using await in a loop will run async function sequentially. Use "
        + "asyncio.gather() to run async functions concurrently."
    )
    VALID = [
        Valid(
            """
            async def async_foo():
                return await async_bar()
            """
        ),
        Valid(
            """
            # await in a loop is fine if it's a test.
            async def async_check_call():
                for _i in range(0, 2):
                    await async_foo()
            """,
            filename="foo/tests/test_foo.py",
        ),
    ]

    INVALID = [
        Invalid(
            """
            async def async_check_call():
                for _i in range(0, 2):
                    await async_foo()
            """,
            line=3,
        ),
        Invalid(
            """
            async def async_check_assignment():
                for _i in range(0, 2):
                    x = await async_foo()
            """,
            line=3,
        ),
        Invalid(
            """
            async def async_check_list_comprehension():
                [await async_foo() for _i in range(0, 2)]
            """,
            line=2,
        ),
        Invalid(
            """
            async def async_check_dict_comprehension_value():
                {_i: await async_foo() for _i in range(0, 2)}
            """,
            line=2,
        ),
        Invalid(
            """
            async def async_check_dict_comprehension_key():
                {await async_foo(_i): _i for _i in range(0, 2)}
            """,
            line=2,
        ),
    ]

    def should_skip_file(self) -> bool:
        return self.context.in_tests

    def visit_Await(self, node: cst.Await) -> None:
        parent = self.context.node_stack[-2]

        if isinstance(parent, (cst.Expr, cst.Assign)) and parent.value is node:
            grand_parent = self.context.node_stack[-5]
            # for and while code block contain IndentBlock and SimpleStatementLine
            if isinstance(grand_parent, (cst.For, cst.While)):
                self.report(node)

        if (
            isinstance(parent, (cst.ListComp, cst.SetComp, cst.GeneratorExp))
            and parent.elt is node
        ):
            self.report(node)

        if isinstance(parent, cst.DictComp) and (
            parent.value is node or parent.key is node
        ):
            self.report(node)



from collections import defaultdict
from typing import Collection, Dict, Iterable, List, Type

from .rule import LintRule, LintRunner
from .types import Config, FileContent, LintViolation


def collect_rules(config: Config) -> Collection[LintRule]:
    # TODO
    from fixit.rules.use_fstring import UseFstringRule

    return [UseFstringRule()]


def collect_violations(
    source: FileContent, rules: Collection[LintRule]
) -> Iterable[LintViolation]:
    # partition rules by LintRunner:
    rules_by_runner: Dict[Type[LintRunner], List[LintRule]] = defaultdict(lambda: [])
    for rule in rules:
        rules_by_runner[rule._runner].append(rule)

    for runner_cls, rules in rules_by_runner.items():
        runner = runner_cls()
        yield from runner.collect_violations(source, rules)
